/******************************************************************************
*                                                                             *
*                                                    *
*                                                     *
*                                                                             *
*   15  2004 .                                                        *
*   (C)                                                         *
*   e-mail       pavelkok@mail.ru                                             *
*   home page    http://kps.nm.ru                                             *
*                                                                             *
*                                                                             *
*   array.h                                                                   *
*                                                              *
*                                                                             *
*******************************************************************************/


#ifndef ARRAY_H
#define ARRAY_H

#define ARRAY_CAPACITY_INC				8
#define SIMPLEARRAY_CAPACITY_INC		4


//     
template <class T>
class Array
{
private:
	T *items;
	int count, capacity;
public:
	Array() { items = NULL; count = 0; capacity = 0; }
	~Array() { delete [] items; }

	int GetCount() { return count; }
	T GetItem(int index);
	T operator [] (int index) { return GetItem(index); }

	void ResetCount() { count = 0; }
	int IndexOf(T item);
	int LastIndexOf(T item);
	void Add(T item);
	void Remove(T item);
};


//     
template <class T>
class SimpleArray
{
private:
	T *items;
	int count;

public:
	SimpleArray()
	{
		items = NULL;
		count = 0;
	}
	~SimpleArray()
	{
		delete [] items;
	}

	int GetCount() { return count; }
	T GetItem(int index);
	T operator [] (int index) { return GetItem(index); }

	int IndexOf(T item);
	int LastIndexOf(T item);
	void Add(T item);
	void AddOnce(T item);
};


template <class T>
T Array<T>::GetItem(int index)
{
	if (index < 0 || index >= count)
		throw L"Index out of range";
	return items[index];
}

template <class T>
int Array<T>::IndexOf(T item)
{
	for (int i = 0; i < count; i++)
		if (items[i] == item)
			return i;
	return -1;
}

template <class T>
int Array<T>::LastIndexOf(T item)
{
	for (int i = count - 1; i >= 0; i--)
		if (items[i] == item)
			return i;
	return -1;
}

template <class T>
void Array<T>::Add(T item)
{
	if (count == capacity)
	{
		capacity += ARRAY_CAPACITY_INC;
		T *new_items = new T[capacity];
		MoveMemory(new_items, items, sizeof(T)*count);
		delete [] items;
		items = new_items;
	}
	items[count++] = item;
}

template <class T>
void Array<T>::Remove(T item)
{
	int i = IndexOf(item);
	if (i >= 0)
	{
		//MoveMemory(items + sizeof(T)*i, items + sizeof(T)*(i + 1), sizeof(T)*(count - i - 1));
		for (int j = i + 1; j < count; j++)
			items[j - 1] = items[j];
		count--;
	}
}


template <class T>
T SimpleArray<T>::GetItem(int index)
{
	if (index < 0 || index >= count)
		throw L"Index out of range";
	return items[index];
}

template <class T>
int SimpleArray<T>::IndexOf(T item)
{
	for (int i = 0; i < count; i++)
		if (items[i] == item)
			return i;
	return -1;
}

template <class T>
int SimpleArray<T>::LastIndexOf(T item)
{
	for (int i = count - 1; i >= 0; i--)
		if (items[i] == item)
			return i;
	return -1;
}

template <class T>
void SimpleArray<T>::Add(T item)
{
	if (count%SIMPLEARRAY_CAPACITY_INC == 0)
	{
		T *new_items = new T[count + SIMPLEARRAY_CAPACITY_INC];
		MoveMemory(new_items, items, sizeof(T)*count);
		delete [] items;
		items = new_items;
	}
	items[count++] = item;
}

template <class T>
void SimpleArray<T>::AddOnce(T item)
{
	if (IndexOf(item) < 0)
		Add(item);
}

#endif
