#include "dlg_materials.h"
#include "dlg_texture.h"
#include "dialog.h"


MaterialArray *editMaterials = NULL;
int currMatIndex;

//    

void LoadMatList(HWND hwnd)
{
	for (int i = 0; i < editMaterials->GetCount(); i++)
	{
		Material *mat = editMaterials->GetItem(i);
		ComboBoxAdd(hwnd, IDC_MATLIST, i, mat->name.GetString());
	}
	SetComboBoxCurrSel(hwnd, IDC_MATLIST, 0);
}

wchar *GetTextureTypeStr(Texture *tex)
{
	if (tex)
		return texInfo[tex->GetType()].typeName;
	else
		return texInfo[ttNone].typeName;
}

//      
void LoadCurrMatParams(HWND hwnd, int currMatIndex)
{
	Material *mat = editMaterials->GetItem(currMatIndex);

	SetEditValue(hwnd, IDC_AMBIENT_R, mat->ambient.r);
	SetEditValue(hwnd, IDC_AMBIENT_G, mat->ambient.g);
	SetEditValue(hwnd, IDC_AMBIENT_B, mat->ambient.b);

	SetEditValue(hwnd, IDC_DIFFUSE_R, mat->diffuse.r);
	SetEditValue(hwnd, IDC_DIFFUSE_G, mat->diffuse.g);
	SetEditValue(hwnd, IDC_DIFFUSE_B, mat->diffuse.b);

	SetEditValue(hwnd, IDC_SPECULAR_R, mat->specular.r);
	SetEditValue(hwnd, IDC_SPECULAR_G, mat->specular.g);
	SetEditValue(hwnd, IDC_SPECULAR_B, mat->specular.b);

	SetEditValue(hwnd, IDC_OPACITY_R, mat->opacity.r);
	SetEditValue(hwnd, IDC_OPACITY_G, mat->opacity.g);
	SetEditValue(hwnd, IDC_OPACITY_B, mat->opacity.b);

	SetEditValue(hwnd, IDC_REFLECTION_R, mat->reflection.r);
	SetEditValue(hwnd, IDC_REFLECTION_G, mat->reflection.g);
	SetEditValue(hwnd, IDC_REFLECTION_B, mat->reflection.b);

	SetEditValue(hwnd, IDC_SPECULAR_LEVEL, mat->specularAmount);
	SetEditValue(hwnd, IDC_GLOSSINESS, mat->specularLevel);
	SetEditValue(hwnd, IDC_REFRACTION_LEVEL, mat->refractionLevel);
	SetEditValue(hwnd, IDC_BUMP_AMOUNT, mat->bumpAmount);

	//      
	SetControlText(hwnd, IDC_AMBIENT_MAP, GetTextureTypeStr(mat->ambientMap));
	SetControlText(hwnd, IDC_DIFFUSE_MAP, GetTextureTypeStr(mat->diffuseMap));
	SetControlText(hwnd, IDC_SPECULAR_MAP, GetTextureTypeStr(mat->specularMap));
	SetControlText(hwnd, IDC_OPACITY_MAP, GetTextureTypeStr(mat->opacityMap));
	SetControlText(hwnd, IDC_REFLECTION_MAP, GetTextureTypeStr(mat->reflectionMap));
	SetControlText(hwnd, IDC_BUMP_MAP, GetTextureTypeStr(mat->bumpMap));
}

void SaveCurrMatParams(HWND hwnd, int currMatIndex)
{
	Material *mat = editMaterials->GetItem(currMatIndex);

	GetEditValue(hwnd, IDC_AMBIENT_R, &mat->ambient.r, 0.0f, 1.0f, 0.0f);
	GetEditValue(hwnd, IDC_AMBIENT_G, &mat->ambient.g, 0.0f, 1.0f, 0.0f);
	GetEditValue(hwnd, IDC_AMBIENT_B, &mat->ambient.b, 0.0f, 1.0f, 0.0f);

	GetEditValue(hwnd, IDC_DIFFUSE_R, &mat->diffuse.r, 0.0f, 1.0f, 0.0f);
	GetEditValue(hwnd, IDC_DIFFUSE_G, &mat->diffuse.g, 0.0f, 1.0f, 0.0f);
	GetEditValue(hwnd, IDC_DIFFUSE_B, &mat->diffuse.b, 0.0f, 1.0f, 0.0f);

	GetEditValue(hwnd, IDC_SPECULAR_R, &mat->specular.r, 0.0f, 1.0f, 0.0f);
	GetEditValue(hwnd, IDC_SPECULAR_G, &mat->specular.g, 0.0f, 1.0f, 0.0f);
	GetEditValue(hwnd, IDC_SPECULAR_B, &mat->specular.b, 0.0f, 1.0f, 0.0f);

	GetEditValue(hwnd, IDC_OPACITY_R, &mat->opacity.r, 0.0f, 1.0f, 0.0f);
	GetEditValue(hwnd, IDC_OPACITY_G, &mat->opacity.g, 0.0f, 1.0f, 0.0f);
	GetEditValue(hwnd, IDC_OPACITY_B, &mat->opacity.b, 0.0f, 1.0f, 0.0f);

	GetEditValue(hwnd, IDC_REFLECTION_R, &mat->reflection.r, 0.0f, 1.0f, 0.0f);
	GetEditValue(hwnd, IDC_REFLECTION_G, &mat->reflection.g, 0.0f, 1.0f, 0.0f);
	GetEditValue(hwnd, IDC_REFLECTION_B, &mat->reflection.b, 0.0f, 1.0f, 0.0f);

	GetEditValue(hwnd, IDC_SPECULAR_LEVEL, &mat->specularAmount, 0.0f, 1.0f, 0.0f);
	GetEditValue(hwnd, IDC_GLOSSINESS, &mat->specularLevel, 0.0f, 1.0f, 0.0f);
	GetEditValue(hwnd, IDC_REFRACTION_LEVEL, &mat->refractionLevel, 0.0f, 100.0f, 1.0f);
	GetEditValue(hwnd, IDC_BUMP_AMOUNT, &mat->bumpAmount, 0.0f, 10.0f, 1.0f);
}

//   
void EditTexture(HWND hwnd, Texture **tex)
{
	editTexture = tex;
	if (DialogBoxParam(hInstance, MAKEINTRESOURCE(IDD_TEXTURE), hwnd, &TextureDialogProc, 0))
	{
		SaveCurrMatParams(hwnd, currMatIndex);
		LoadCurrMatParams(hwnd, currMatIndex);
	}
}

//     
void DeleteMaterials(MaterialArray *mats)
{
	if (mats)
	{
		for (int i = 0; i < mats->GetCount(); i++)
			delete mats->GetItem(i);
		mats->ResetCount();
	}
}

void DeleteMaterialsArray(MaterialArray **materials)
{
	DeleteMaterials(*materials);
	delete *materials;
	*materials = NULL;
}

//    
void LoadEditMaterials(MaterialArray *materials)
{
	DeleteMaterialsArray(&editMaterials);
	editMaterials = new MaterialArray();

	for (int i = 0; i < materials->GetCount(); i++)
		editMaterials->Add(materials->GetItem(i)->Copy());
}

//    
void SaveEditMaterials(MaterialArray *materials)
{
	DeleteMaterials(materials);
	for (int i = 0; i < editMaterials->GetCount(); i++)
		materials->Add(editMaterials->GetItem(i)->Copy());
}


// Cancel
void ClearEditMaterials()
{
	DeleteMaterialsArray(&editMaterials);
}

// OK
void ClearAndSaveClearEditMaterials()
{
	SaveEditMaterials(&scene->materials);
	DeleteMaterialsArray(&editMaterials);
}


int CALLBACK MaterialDialogProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_INITDIALOG:
		LoadEditMaterials(&scene->materials);
		LoadMatList(hwnd);
		currMatIndex = 0;
		LoadCurrMatParams(hwnd, currMatIndex);
		return TRUE;

	case WM_CLOSE:
		ClearEditMaterials();
		EndDialog(hwnd, 0);
		return TRUE;

	case WM_COMMAND:
	{
		int controlID = LOWORD(wParam);
		int code = HIWORD(wParam);
		HWND hwndCtrl = (HWND)lParam;

		switch (controlID)
		{
		case IDOK:
			SaveCurrMatParams(hwnd, currMatIndex);
			ClearAndSaveClearEditMaterials();
			EndDialog(hwnd, 1);
			break;
		case IDCANCEL:
			ClearEditMaterials();
			EndDialog(hwnd, 0);
			break;

		case IDC_MATLIST:
			if (code == CBN_SELCHANGE)
			{
				SaveCurrMatParams(hwnd, currMatIndex);
				currMatIndex = GetComboBoxCurrSel(hwnd, IDC_MATLIST);
				LoadCurrMatParams(hwnd, currMatIndex);
			}
			break;

		case IDC_AMBIENT_MAP:
			EditTexture(hwnd, &(editMaterials->GetItem(currMatIndex)->ambientMap));
			break;
		case IDC_DIFFUSE_MAP:
			EditTexture(hwnd, &(editMaterials->GetItem(currMatIndex)->diffuseMap));
			break;
		case IDC_SPECULAR_MAP:
			EditTexture(hwnd, &(editMaterials->GetItem(currMatIndex)->specularMap));
			break;
		case IDC_OPACITY_MAP:
			EditTexture(hwnd, &(editMaterials->GetItem(currMatIndex)->opacityMap));
			break;
		case IDC_REFLECTION_MAP:
			EditTexture(hwnd, &(editMaterials->GetItem(currMatIndex)->reflectionMap));
			break;
		case IDC_BUMP_MAP:
			EditTexture(hwnd, &(editMaterials->GetItem(currMatIndex)->bumpMap));
			break;
		}
		return 0;
	}

	case WM_NOTIFY:
	{
		LPNMHDR pnmh = (LPNMHDR)lParam;
		switch (pnmh->idFrom)
		{
		case IDC_MATLIST:
			break;
		}
		return 0;
	}
	}
	return 0;
}
