#include "dlg_materials.h"
#include "dlg_texture.h"
#include "dialog.h"


TextureType currType;
Texture **editTexture;
HWND hwndDlg[DlgTextureInfoCount];

DlgTextureInfo texInfo[DlgTextureInfoCount] = 
{
	// ttNone
	{
		0, 0, L"None"
	},
	// ttBitmap
	{
		IDD_TEX_BITMAP, &TextureBitmapDialogProc, L"Bitmap"
	},
	// ttChessboard
	{
		IDD_TEX_CHECKER, &TextureCheckerDialogProc, L"Checker"
	},
	// ttPerlinNoise
	{
		IDD_TEX_PERLINNOISE, &TexturePerlinNoiseDialogProc, L"Perlin Noise"
	}
};


void LoadTextureTypeList(HWND hwnd)
{
	for (int i = 0; i < DlgTextureInfoCount; i++)
		ComboBoxAdd(hwnd, IDC_TEXTURE_TYPELIST, i, texInfo[i].typeName);
}

void SetTextureType(HWND hwnd, TextureType newType)
{
	if (hwndDlg[currType])
		ShowWindow(hwndDlg[currType], SW_HIDE);
	currType = newType;
	SetWindowPos(hwndDlg[currType], HWND_TOP, 
		childDlgX, childDlgY, childDlgCX, childDlgCY + 20, SWP_SHOWWINDOW);

	SetComboBoxCurrSel(hwnd, IDC_TEXTURE_TYPELIST, currType);
}

//    
void LoadTexture(HWND hwnd, Texture *texture)
{
	if (texture)
	{
		HWND hwndDialog = hwndDlg[texture->GetType()];

		SetEditValue(hwnd, IDC_OFFSET_X, texture->offset.x);
		SetEditValue(hwnd, IDC_OFFSET_Y, texture->offset.y);
		SetEditValue(hwnd, IDC_OFFSET_Z, texture->offset.z);

		SetEditValue(hwnd, IDC_TILING_X, texture->tiling.x);
		SetEditValue(hwnd, IDC_TILING_Y, texture->tiling.y);
		SetEditValue(hwnd, IDC_TILING_Z, texture->tiling.z);

		SetCheckBoxState(hwnd, IDC_INVERTCOLORS, texture->invert);

		switch (texture->GetType())
		{
		case ttBitmap:
		{
			TextureBitmap *tex = (TextureBitmap*)texture;
			SetControlText(hwndDialog, IDC_TEX_BITMAP_FILENAME, tex->fileName.GetString());
			break;
		}
		case ttChessboard:
		{
			TextureChessboard *tex = (TextureChessboard*)texture;

			SetEditValue(hwndDialog, IDC_TEX_CHECKER_COLOR1_R, tex->evenColor.r);
			SetEditValue(hwndDialog, IDC_TEX_CHECKER_COLOR1_G, tex->evenColor.g);
			SetEditValue(hwndDialog, IDC_TEX_CHECKER_COLOR1_B, tex->evenColor.b);

			SetEditValue(hwndDialog, IDC_TEX_CHECKER_COLOR2_R, tex->oddColor.r);
			SetEditValue(hwndDialog, IDC_TEX_CHECKER_COLOR2_G, tex->oddColor.g);
			SetEditValue(hwndDialog, IDC_TEX_CHECKER_COLOR2_B, tex->oddColor.b);
			break;
		}
		case ttPerlinNoise:
		{
			TexturePerlinNoise *tex = (TexturePerlinNoise*)texture;

			SetEditValue(hwndDialog, IDC_TEX_PERLINNOISE_COLOR1_R, tex->baseColor.r);
			SetEditValue(hwndDialog, IDC_TEX_PERLINNOISE_COLOR1_G, tex->baseColor.g);
			SetEditValue(hwndDialog, IDC_TEX_PERLINNOISE_COLOR1_B, tex->baseColor.b);

			SetEditValue(hwndDialog, IDC_TEX_PERLINNOISE_COLOR2_R, tex->noiseColor.r);
			SetEditValue(hwndDialog, IDC_TEX_PERLINNOISE_COLOR2_G, tex->noiseColor.g);
			SetEditValue(hwndDialog, IDC_TEX_PERLINNOISE_COLOR2_B, tex->noiseColor.b);

			SetEditValue(hwndDialog, IDC_TEX_PERLINNOISE_AMPLITUDE, tex->amplitude);
			SetEditValue(hwndDialog, IDC_TEX_PERLINNOISE_FREQUENCY, tex->frequency);
			SetEditValue(hwndDialog, IDC_TEX_PERLINNOISE_PERSISTENCE, tex->persistence);
			SetEditValue(hwndDialog, IDC_TEX_PERLINNOISE_OCTAVES, (REAL)tex->octaves);
			break;
		}
		}
		SetTextureType(hwnd, texture->GetType());
	}
	else
	{
		SetTextureType(hwnd, ttNone);

		SetEditValue(hwnd, IDC_OFFSET_X, 0.0f);
		SetEditValue(hwnd, IDC_OFFSET_Y, 0.0f);
		SetEditValue(hwnd, IDC_OFFSET_Z, 0.0f);

		SetEditValue(hwnd, IDC_TILING_X, 1.0f);
		SetEditValue(hwnd, IDC_TILING_Y, 1.0f);
		SetEditValue(hwnd, IDC_TILING_Z, 1.0f);

		SetCheckBoxState(hwnd, IDC_INVERTCOLORS, false);
	}
}

void SaveTexture(HWND hwnd, Texture **texture)
{
	if (*texture)
	{
		delete *texture;
		*texture = NULL;
	}

	HWND hwndDialog = hwndDlg[currType];
	const int bufLen = 256;
	wchar buf[bufLen];

	switch (currType)
	{
	case ttBitmap:
	{
		GetControlText(hwndDialog, IDC_TEX_BITMAP_FILENAME, buf, bufLen);
		TextureBitmap* tex = new TextureBitmap(buf);
		*texture = tex;
		break;
	}

	case ttChessboard:
	{
		TextureChessboard* tex = new TextureChessboard();
		*texture = tex;

		GetEditValue(hwndDialog, IDC_TEX_CHECKER_COLOR1_R, &tex->evenColor.r, 0.0f, 1.0f, 0.0f);
		GetEditValue(hwndDialog, IDC_TEX_CHECKER_COLOR1_G, &tex->evenColor.g, 0.0f, 1.0f, 0.0f);
		GetEditValue(hwndDialog, IDC_TEX_CHECKER_COLOR1_B, &tex->evenColor.b, 0.0f, 1.0f, 0.0f);

		GetEditValue(hwndDialog, IDC_TEX_CHECKER_COLOR2_R, &tex->oddColor.r, 0.0f, 1.0f, 1.0f);
		GetEditValue(hwndDialog, IDC_TEX_CHECKER_COLOR2_G, &tex->oddColor.g, 0.0f, 1.0f, 1.0f);
		GetEditValue(hwndDialog, IDC_TEX_CHECKER_COLOR2_B, &tex->oddColor.b, 0.0f, 1.0f, 1.0f);
		break;
	}
	case ttPerlinNoise:
	{
		REAL amplitude, frequency, persistence, octaves;
		GetEditValue(hwndDialog, IDC_TEX_PERLINNOISE_AMPLITUDE, &amplitude, 0.0f, 1000.0f, 1.0f);
		GetEditValue(hwndDialog, IDC_TEX_PERLINNOISE_FREQUENCY, &frequency, 0.0f, 1000.0f, 4.0f);
		GetEditValue(hwndDialog, IDC_TEX_PERLINNOISE_PERSISTENCE, &persistence, 0.0f, 1.0f, 0.5f);
		GetEditValue(hwndDialog, IDC_TEX_PERLINNOISE_OCTAVES, &octaves, 0.0f, 256.0f, 8.0f);

		TexturePerlinNoise *tex = new TexturePerlinNoise(amplitude, frequency, persistence, (int)octaves);
		*texture = tex;

		GetEditValue(hwndDialog, IDC_TEX_PERLINNOISE_COLOR1_R, &tex->baseColor.r, 0.0f, 1.0f, 0.0f);
		GetEditValue(hwndDialog, IDC_TEX_PERLINNOISE_COLOR1_G, &tex->baseColor.g, 0.0f, 1.0f, 0.0f);
		GetEditValue(hwndDialog, IDC_TEX_PERLINNOISE_COLOR1_B, &tex->baseColor.b, 0.0f, 1.0f, 0.0f);

		GetEditValue(hwndDialog, IDC_TEX_PERLINNOISE_COLOR2_R, &tex->noiseColor.r, 0.0f, 1.0f, 1.0f);
		GetEditValue(hwndDialog, IDC_TEX_PERLINNOISE_COLOR2_G, &tex->noiseColor.g, 0.0f, 1.0f, 1.0f);
		GetEditValue(hwndDialog, IDC_TEX_PERLINNOISE_COLOR2_B, &tex->noiseColor.b, 0.0f, 1.0f, 1.0f);

		break;
	}
	}

	if (*texture)
	{
		Texture *tex = *texture;

		GetEditValue(hwnd, IDC_OFFSET_X, &tex->offset.x, -1000.0f, 1000.0f, 0.0f);
		GetEditValue(hwnd, IDC_OFFSET_Y, &tex->offset.y, -1000.0f, 1000.0f, 0.0f);
		GetEditValue(hwnd, IDC_OFFSET_Z, &tex->offset.z, -1000.0f, 1000.0f, 0.0f);

		GetEditValue(hwnd, IDC_TILING_X, &tex->tiling.x, -1000.0f, 1000.0f, 1.0f);
		GetEditValue(hwnd, IDC_TILING_Y, &tex->tiling.y, -1000.0f, 1000.0f, 1.0f);
		GetEditValue(hwnd, IDC_TILING_Z, &tex->tiling.z, -1000.0f, 1000.0f, 1.0f);

		GetCheckBoxState(hwnd, IDC_INVERTCOLORS, &tex->invert);
	}
}


int CALLBACK TextureBitmapDialogProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_INITDIALOG:
		SetControlText(hwnd, IDC_TEX_BITMAP_FILENAME, L"");
		return TRUE;
	}
	return 0;
}

int CALLBACK TextureCheckerDialogProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_INITDIALOG:
		SetEditValue(hwnd, IDC_TEX_CHECKER_COLOR1_R, 0.0f);
		SetEditValue(hwnd, IDC_TEX_CHECKER_COLOR1_G, 0.0f);
		SetEditValue(hwnd, IDC_TEX_CHECKER_COLOR1_B, 0.0f);

		SetEditValue(hwnd, IDC_TEX_CHECKER_COLOR2_R, 1.0f);
		SetEditValue(hwnd, IDC_TEX_CHECKER_COLOR2_G, 1.0f);
		SetEditValue(hwnd, IDC_TEX_CHECKER_COLOR2_B, 1.0f);

		SetEditValue(hwnd, IDC_TEX_CHECKER_REPEAT_X, 2);
		SetEditValue(hwnd, IDC_TEX_CHECKER_REPEAT_Y, 2);
		SetEditValue(hwnd, IDC_TEX_CHECKER_REPEAT_Z, 2);
		return TRUE;
	}
	return 0;
}

int CALLBACK TexturePerlinNoiseDialogProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_INITDIALOG:
		SetEditValue(hwnd, IDC_TEX_PERLINNOISE_COLOR1_R, 0.0f);
		SetEditValue(hwnd, IDC_TEX_PERLINNOISE_COLOR1_G, 0.0f);
		SetEditValue(hwnd, IDC_TEX_PERLINNOISE_COLOR1_B, 0.0f);

		SetEditValue(hwnd, IDC_TEX_PERLINNOISE_COLOR2_R, 1.0f);
		SetEditValue(hwnd, IDC_TEX_PERLINNOISE_COLOR2_G, 1.0f);
		SetEditValue(hwnd, IDC_TEX_PERLINNOISE_COLOR2_B, 1.0f);

		SetEditValue(hwnd, IDC_TEX_PERLINNOISE_AMPLITUDE, 1.0);
		SetEditValue(hwnd, IDC_TEX_PERLINNOISE_FREQUENCY, 4.0);
		SetEditValue(hwnd, IDC_TEX_PERLINNOISE_PERSISTENCE, 0.5);
		SetEditValue(hwnd, IDC_TEX_PERLINNOISE_OCTAVES, (REAL)8);
		return TRUE;
	}
	return 0;
}


int CALLBACK TextureDialogProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_INITDIALOG:
	{
		currType = ttNone;
		LoadTextureTypeList(hwnd);

		for (int i = 1; i < DlgTextureInfoCount; i++)
			hwndDlg[i] = CreateDialogParam(
				hInstance, MAKEINTRESOURCE(texInfo[i].id), hwnd, texInfo[i].proc, 0);

		LoadTexture(hwnd, *editTexture);

		return TRUE;
	}

	case WM_CLOSE:
		EndDialog(hwnd, 0);
		return TRUE;

	case WM_COMMAND:
	{
		int controlID = LOWORD(wParam);
		int code = HIWORD(wParam);
		HWND hwndCtrl = (HWND)lParam;

		switch (controlID)
		{
		case IDOK:
			SaveTexture(hwnd, editTexture);
			EndDialog(hwnd, 1);
			break;
		case IDCANCEL:
			EndDialog(hwnd, 0);
			break;
		case IDC_TEXTURE_TYPELIST:
			if (code == CBN_SELCHANGE)
			{
				int index = (int)SendMessage(hwndCtrl, CB_GETCURSEL, 0, 0);
				SetTextureType(hwnd, (TextureType)index);
			}
		}
		return 0;
	}
	}
	return 0;
}
