#include "impt_ase.h"


LoaderASE::LoaderASE(wchar *fileName, PScene scene): stream(fileName, fsmOpenExisting), reader(&stream, 1024)
{
	nLine = 1;
	this->scene = scene;

	CreateTable();

	NextChar();
	NextToken();

	try
	{
		Parse();
	}
	catch (wchar *msgEx)
	{
		wchar msg[256];
		wsprintf(msg, 
			L"Error load ASE\nFile: \"%s\"\nException: %s\nAt line %d",
			fileName, msgEx, nLine);
		throw msg;
	}
}

LoaderASE::~LoaderASE()
{
	if (hash)
		delete hash;
}

void LoaderASE::Warning(wchar *msg)
{
	MessageBox(NULL, msg, L"Warning!", MB_ICONWARNING);
}

void LoaderASE::CreateTable()
{
	hash = new HashTable();

	hash->Add(ST_SCENE, ID_SCENE);
	hash->Add(ST_GEOMETRY, ID_GEOMETRY);
	hash->Add(ST_SHAPE, ID_SHAPE);
	hash->Add(ST_CAMERA, ID_CAMERA);
	hash->Add(ST_LIGHT, ID_LIGHT);
	hash->Add(ST_HELPER, ID_HELPER);
	hash->Add(ST_MATERIAL_LIST, ID_MATERIAL_LIST);
	hash->Add(ST_GROUP, ID_GROUP);
	hash->Add(ST_NODE_TM, ID_NODE_TM);
	hash->Add(ST_NODE_NAME, ID_NODE_NAME);
	hash->Add(ST_NODE_PARENT, ID_NODE_PARENT);
	hash->Add(ST_PROP_MOTIONBLUR, ID_PROP_MOTIONBLUR);
	hash->Add(ST_PROP_CASTSHADOW, ID_PROP_CASTSHADOW);
	hash->Add(ST_PROP_RECVSHADOW, ID_PROP_RECVSHADOW);
	hash->Add(ST_MESH, ID_MESH);
	hash->Add(ST_MESH_NORMALS, ID_MESH_NORMALS);
	hash->Add(ST_MESH_NUMVERTEX, ID_MESH_NUMVERTEX);
	hash->Add(ST_MESH_NUMFACES, ID_MESH_NUMFACES);
	hash->Add(ST_MESH_VERTEX_LIST, ID_MESH_VERTEX_LIST);
	hash->Add(ST_MESH_VERTEX, ID_MESH_VERTEX);
	hash->Add(ST_MESH_FACE_LIST, ID_MESH_FACE_LIST);

	hash->Add(ST_MESH_FACE_A, ID_MESH_FACE_A);
	hash->Add(ST_MESH_FACE_B, ID_MESH_FACE_B);
	hash->Add(ST_MESH_FACE_C, ID_MESH_FACE_C);
	hash->Add(ST_MESH_FACE_AB, ID_MESH_FACE_AB);
	hash->Add(ST_MESH_FACE_BC, ID_MESH_FACE_BC);
	hash->Add(ST_MESH_FACE_CA, ID_MESH_FACE_CA);

	hash->Add(ST_MESH_FACE, ID_MESH_FACE);
	hash->Add(ST_MESH_SMOOTHING, ID_MESH_SMOOTHING);
	hash->Add(ST_MESH_MTLID, ID_MESH_MTLID);
	hash->Add(ST_MESH_NUMTVERTEX, ID_MESH_NUMTVERTEX);
	hash->Add(ST_MESH_NUMTVFACES, ID_MESH_NUMTVFACES);
	hash->Add(ST_MESH_TVERTLIST, ID_MESH_TVERTLIST);
	hash->Add(ST_MESH_TVERT, ID_MESH_TVERT);
	hash->Add(ST_MESH_TFACELIST, ID_MESH_TFACELIST);
	hash->Add(ST_MESH_TFACE, ID_MESH_TFACE);
	hash->Add(ST_MESH_NUMCVERTEX, ID_MESH_NUMCVERTEX);
	hash->Add(ST_MESH_NUMCVFACES, ID_MESH_NUMCVFACES);
	hash->Add(ST_MESH_CVERTLIST, ID_MESH_CVERTLIST);
	hash->Add(ST_MESH_VERTCOL, ID_MESH_VERTCOL);
	hash->Add(ST_MESH_CFACELIST, ID_MESH_CFACELIST);
	hash->Add(ST_MESH_CFACE, ID_MESH_CFACE);
	hash->Add(ST_MESH_FACEMAPLIST, ID_MESH_FACEMAPLIST);
	hash->Add(ST_MESH_FACEMAP, ID_MESH_FACEMAP);
	hash->Add(ST_MESH_FACEVERT, ID_MESH_FACEVERT);
	hash->Add(ST_MESH_FACENORMAL, ID_MESH_FACENORMAL);
	hash->Add(ST_MESH_VERTEXNORMAL, ID_MESH_VERTEXNORMAL);
	hash->Add(ST_MESH_ANIMATION, ID_MESH_ANIMATION);
	hash->Add(ST_SHAPE_LINECOUNT, ID_SHAPE_LINECOUNT);
	hash->Add(ST_SHAPE_LINE, ID_SHAPE_LINE);
	hash->Add(ST_SHAPE_VERTEX_KNOT, ID_SHAPE_VERTEX_KNOT);
	hash->Add(ST_SHAPE_VERTEX_INTERP, ID_SHAPE_VERTEX_INTERP);
	hash->Add(ST_SHAPE_VERTEXCOUNT, ID_SHAPE_VERTEXCOUNT);
	hash->Add(ST_SHAPE_CLOSED, ID_SHAPE_CLOSED);
	hash->Add(ST_LIGHT_SETTINGS, ID_LIGHT_SETTINGS);
	hash->Add(ST_LIGHT_TYPE, ID_LIGHT_TYPE);
	hash->Add(ST_LIGHT_COLOR, ID_LIGHT_COLOR);
	hash->Add(ST_LIGHT_INTENS, ID_LIGHT_INTENS);
	hash->Add(ST_LIGHT_HOTSPOT, ID_LIGHT_HOTSPOT);
	hash->Add(ST_LIGHT_FALLOFF, ID_LIGHT_FALLOFF);
	hash->Add(ST_LIGHT_ATTNSTART, ID_LIGHT_ATTNSTART);
	hash->Add(ST_LIGHT_ATTNEND, ID_LIGHT_ATTNEND);
	hash->Add(ST_LIGHT_ASPECT, ID_LIGHT_ASPECT);
	hash->Add(ST_LIGHT_SHADOWS, ID_LIGHT_SHADOWS);
	hash->Add(ST_LIGHT_USELIGHT, ID_LIGHT_USELIGHT);
	hash->Add(ST_LIGHT_SPOTSHAPE, ID_LIGHT_SPOTSHAPE);
	hash->Add(ST_LIGHT_TDIST, ID_LIGHT_TDIST);
	hash->Add(ST_LIGHT_MAPBIAS, ID_LIGHT_MAPBIAS);
	hash->Add(ST_LIGHT_MAPRANGE, ID_LIGHT_MAPRANGE);
	hash->Add(ST_LIGHT_MAPSIZE, ID_LIGHT_MAPSIZE);
	hash->Add(ST_LIGHT_RAYBIAS, ID_LIGHT_RAYBIAS);
	hash->Add(ST_LIGHT_USEGLOBAL, ID_LIGHT_USEGLOBAL);
	hash->Add(ST_LIGHT_ABSMAPBIAS, ID_LIGHT_ABSMAPBIAS);
	hash->Add(ST_LIGHT_OVERSHOOT, ID_LIGHT_OVERSHOOT);
	hash->Add(ST_LIGHT_EXCLUSIONLIST, ID_LIGHT_EXCLUSIONLIST);
	hash->Add(ST_LIGHT_NUMEXCLUDED, ID_LIGHT_NUMEXCLUDED);
	hash->Add(ST_LIGHT_EXCLUDED, ID_LIGHT_EXCLUDED);
	hash->Add(ST_LIGHT_EXCLINCLUDE, ID_LIGHT_EXCLINCLUDE);
	hash->Add(ST_LIGHT_EXCL_AFFECT_ILLUM, ID_LIGHT_EXCL_AFFECT_ILLUM);
	hash->Add(ST_LIGHT_EXCL_AFFECT_SHAD, ID_LIGHT_EXCL_AFFECT_SHAD);
	hash->Add(ST_LIGHT_ANIMATION, ID_LIGHT_ANIMATION);
	hash->Add(ST_LIGHT_TYPE_OMNI, ID_LIGHT_TYPE_OMNI);
	hash->Add(ST_LIGHT_TYPE_TARG, ID_LIGHT_TYPE_TARG);
	hash->Add(ST_LIGHT_TYPE_DIR, ID_LIGHT_TYPE_DIR);
	hash->Add(ST_LIGHT_TYPE_FREE, ID_LIGHT_TYPE_FREE);
	hash->Add(ST_LIGHT_SHAD_OFF, ID_LIGHT_SHAD_OFF);
	hash->Add(ST_LIGHT_SHAD_MAP, ID_LIGHT_SHAD_MAP);
	hash->Add(ST_LIGHT_SHAD_RAY, ID_LIGHT_SHAD_RAY);
	hash->Add(ST_LIGHT_SHAPE_CIRC, ID_LIGHT_SHAPE_CIRC);
	hash->Add(ST_LIGHT_SHAPE_RECT, ID_LIGHT_SHAPE_RECT);
	hash->Add(ST_CAMERA_SETTINGS, ID_CAMERA_SETTINGS);
	hash->Add(ST_CAMERA_HITHER, ID_CAMERA_HITHER);
	hash->Add(ST_CAMERA_YON, ID_CAMERA_YON);
	hash->Add(ST_CAMERA_NEAR, ID_CAMERA_NEAR);
	hash->Add(ST_CAMERA_FAR, ID_CAMERA_FAR);
	hash->Add(ST_CAMERA_FOV, ID_CAMERA_FOV);
	hash->Add(ST_CAMERA_TDIST, ID_CAMERA_TDIST);
	hash->Add(ST_CAMERA_ANIMATION, ID_CAMERA_ANIMATION);
	hash->Add(ST_CAMERA_TYPE, ID_CAMERA_TYPE);
	hash->Add(ST_CAMERATYPE_TARGET, ID_CAMERATYPE_TARGET);
	hash->Add(ST_CAMERATYPE_FREE, ID_CAMERATYPE_FREE);
	hash->Add(ST_HELPER_CLASS, ID_HELPER_CLASS);
	hash->Add(ST_CONTROL_POINT3_TCB, ID_CONTROL_POINT3_TCB);
	hash->Add(ST_CONTROL_POINT3_BEZIER, ID_CONTROL_POINT3_BEZIER);
	hash->Add(ST_CONTROL_COLOR_BEZIER, ID_CONTROL_COLOR_BEZIER);
	hash->Add(ST_CONTROL_POINT3_SAMPLE, ID_CONTROL_POINT3_SAMPLE);
	hash->Add(ST_CONTROL_FLOAT_TCB, ID_CONTROL_FLOAT_TCB);
	hash->Add(ST_CONTROL_FLOAT_BEZIER, ID_CONTROL_FLOAT_BEZIER);
	hash->Add(ST_CONTROL_FLOAT_LINEAR, ID_CONTROL_FLOAT_LINEAR);
	hash->Add(ST_CONTROL_FLOAT_SAMPLE, ID_CONTROL_FLOAT_SAMPLE);
	hash->Add(ST_POS_TRACK, ID_POS_TRACK);
	hash->Add(ST_ROT_TRACK, ID_ROT_TRACK);
	hash->Add(ST_SCALE_TRACK, ID_SCALE_TRACK);
	hash->Add(ST_POS_SAMPLE, ID_POS_SAMPLE);
	hash->Add(ST_ROT_SAMPLE, ID_ROT_SAMPLE);
	hash->Add(ST_SCALE_SAMPLE, ID_SCALE_SAMPLE);
	hash->Add(ST_POS_KEY, ID_POS_KEY);
	hash->Add(ST_ROT_KEY, ID_ROT_KEY);
	hash->Add(ST_SCALE_KEY, ID_SCALE_KEY);
	hash->Add(ST_POINT3_KEY, ID_POINT3_KEY);
	hash->Add(ST_FLOAT_KEY, ID_FLOAT_KEY);
	hash->Add(ST_TCB_POINT3_KEY, ID_TCB_POINT3_KEY);
	hash->Add(ST_TCB_FLOAT_KEY, ID_TCB_FLOAT_KEY);
	hash->Add(ST_TCB_POS_KEY, ID_TCB_POS_KEY);
	hash->Add(ST_TCB_ROT_KEY, ID_TCB_ROT_KEY);
	hash->Add(ST_TCB_SCALE_KEY, ID_TCB_SCALE_KEY);
	hash->Add(ST_BEZIER_FLOAT_KEY, ID_BEZIER_FLOAT_KEY);
	hash->Add(ST_BEZIER_POINT3_KEY, ID_BEZIER_POINT3_KEY);
	hash->Add(ST_BEZIER_POS_KEY, ID_BEZIER_POS_KEY);
	hash->Add(ST_BEZIER_SCALE_KEY, ID_BEZIER_SCALE_KEY);
	hash->Add(ST_CONTROL_POS_LINEAR, ID_CONTROL_POS_LINEAR);
	hash->Add(ST_CONTROL_POS_TCB, ID_CONTROL_POS_TCB);
	hash->Add(ST_CONTROL_POS_BEZIER, ID_CONTROL_POS_BEZIER);
	hash->Add(ST_CONTROL_ROT_LINEAR, ID_CONTROL_ROT_LINEAR);
	hash->Add(ST_CONTROL_ROT_TCB, ID_CONTROL_ROT_TCB);
	hash->Add(ST_CONTROL_ROT_BEZIER, ID_CONTROL_ROT_BEZIER);
	hash->Add(ST_CONTROL_SCALE_LINEAR , ID_CONTROL_SCALE_LINEAR );
	hash->Add(ST_CONTROL_SCALE_TCB, ID_CONTROL_SCALE_TCB);
	hash->Add(ST_CONTROL_SCALE_BEZIER, ID_CONTROL_SCALE_BEZIER);
	hash->Add(ST_IKTERMINATOR, ID_IKTERMINATOR);
	hash->Add(ST_IKROT_PINNED, ID_IKROT_PINNED);
	hash->Add(ST_IKPOS_PINNED, ID_IKPOS_PINNED);
	hash->Add(ST_IKJOINT, ID_IKJOINT);
	hash->Add(ST_IKTYPE, ID_IKTYPE);
	hash->Add(ST_IKDOF, ID_IKDOF);
	hash->Add(ST_IKXACTIVE, ID_IKXACTIVE);
	hash->Add(ST_IKYACTIVE, ID_IKYACTIVE);
	hash->Add(ST_IKZACTIVE, ID_IKZACTIVE);
	hash->Add(ST_IKXLIMITED, ID_IKXLIMITED);
	hash->Add(ST_IKYLIMITED, ID_IKYLIMITED);
	hash->Add(ST_IKZLIMITED, ID_IKZLIMITED);
	hash->Add(ST_IKXEASE, ID_IKXEASE);
	hash->Add(ST_IKYEASE, ID_IKYEASE);
	hash->Add(ST_IKZEASE, ID_IKZEASE);
	hash->Add(ST_IKLIMITEXACT, ID_IKLIMITEXACT);
	hash->Add(ST_IKJOINTINFO, ID_IKJOINTINFO);
	hash->Add(ST_IKTYPEPOS, ID_IKTYPEPOS);
	hash->Add(ST_IKTYPEROT, ID_IKTYPEROT);
	hash->Add(ST_WIRECOLOR, ID_WIRECOLOR);
	hash->Add(ST_MATERIAL, ID_MATERIAL);
	hash->Add(ST_MATERIAL_COUNT, ID_MATERIAL_COUNT);
	hash->Add(ST_MATERIAL_REF, ID_MATERIAL_REF);
	hash->Add(ST_NUMSUBMTLS, ID_NUMSUBMTLS);
	hash->Add(ST_SUBMATERIAL, ID_SUBMATERIAL);
	hash->Add(ST_MATNAME, ID_MATNAME);
	hash->Add(ST_MATCLASS, ID_MATCLASS);
	hash->Add(ST_MAT_SHADE_CONST, ID_MAT_SHADE_CONST);
	hash->Add(ST_MAT_SHADE_PHONG, ID_MAT_SHADE_PHONG);
	hash->Add(ST_MAT_SHADE_METAL, ID_MAT_SHADE_METAL);
	hash->Add(ST_MAT_SHADE_BLINN, ID_MAT_SHADE_BLINN);
	hash->Add(ST_MAT_SHADE_OTHER, ID_MAT_SHADE_OTHER);
	hash->Add(ST_MAP_XPTYPE_FLT, ID_MAP_XPTYPE_FLT);
	hash->Add(ST_MAP_XPTYPE_SUB, ID_MAP_XPTYPE_SUB);
	hash->Add(ST_MAP_XPTYPE_ADD, ID_MAP_XPTYPE_ADD);
	hash->Add(ST_MAP_XPTYPE_OTH, ID_MAP_XPTYPE_OTH);
	hash->Add(ST_BMP_FILT_PYR, ID_BMP_FILT_PYR);
	hash->Add(ST_BMP_FILT_SAT, ID_BMP_FILT_SAT);
	hash->Add(ST_BMP_FILT_NONE, ID_BMP_FILT_NONE);
	hash->Add(ST_FALLOFF_OUT, ID_FALLOFF_OUT);
	hash->Add(ST_FALLOFF_IN, ID_FALLOFF_IN);
	hash->Add(ST_MAPTYPE_EXP, ID_MAPTYPE_EXP);
	hash->Add(ST_MAPTYPE_SPH, ID_MAPTYPE_SPH);
	hash->Add(ST_MAPTYPE_CYL, ID_MAPTYPE_CYL);
	hash->Add(ST_MAPTYPE_SHR, ID_MAPTYPE_SHR);
	hash->Add(ST_MAPTYPE_SCR, ID_MAPTYPE_SCR);
	hash->Add(ST_AMBIENT, ID_AMBIENT);
	hash->Add(ST_DIFFUSE, ID_DIFFUSE);
	hash->Add(ST_SPECULAR, ID_SPECULAR);
	hash->Add(ST_SHINE, ID_SHINE);
	hash->Add(ST_SHINE_STRENGTH, ID_SHINE_STRENGTH);
	hash->Add(ST_TRANSPARENCY, ID_TRANSPARENCY);
	hash->Add(ST_WIRESIZE, ID_WIRESIZE);
	hash->Add(ST_SHADING, ID_SHADING);
	hash->Add(ST_XP_FALLOFF, ID_XP_FALLOFF);
	hash->Add(ST_SELFILLUM, ID_SELFILLUM);
	hash->Add(ST_TWOSIDED, ID_TWOSIDED);
	hash->Add(ST_WIRE, ID_WIRE);
	hash->Add(ST_WIREUNITS, ID_WIREUNITS);
	hash->Add(ST_FALLOFF, ID_FALLOFF);
	hash->Add(ST_FACEMAP, ID_FACEMAP);
	hash->Add(ST_SOFTEN, ID_SOFTEN);
	hash->Add(ST_XP_TYPE, ID_XP_TYPE);
	hash->Add(ST_TEXNAME, ID_TEXNAME);
	hash->Add(ST_TEXCLASS, ID_TEXCLASS);
	hash->Add(ST_TEXSUBNO, ID_TEXSUBNO);
	hash->Add(ST_TEXAMOUNT, ID_TEXAMOUNT);
	hash->Add(ST_BITMAP, ID_BITMAP);
	hash->Add(ST_TEX_INVERT, ID_TEX_INVERT);
	hash->Add(ST_BMP_FILTER, ID_BMP_FILTER);
	hash->Add(ST_MAPTYPE, ID_MAPTYPE);
	hash->Add(ST_U_OFFSET, ID_U_OFFSET);
	hash->Add(ST_V_OFFSET, ID_V_OFFSET);
	hash->Add(ST_U_TILING, ID_U_TILING);
	hash->Add(ST_V_TILING, ID_V_TILING);
	hash->Add(ST_ANGLE, ID_ANGLE);
	hash->Add(ST_BLUR, ID_BLUR);
	hash->Add(ST_BLUR_OFFSET, ID_BLUR_OFFSET);
	hash->Add(ST_NOISE_AMT, ID_NOISE_AMT);
	hash->Add(ST_NOISE_SIZE, ID_NOISE_SIZE);
	hash->Add(ST_NOISE_LEVEL, ID_NOISE_LEVEL);
	hash->Add(ST_NOISE_PHASE, ID_NOISE_PHASE);
	hash->Add(ST_MAP_GENERIC, ID_MAP_GENERIC);
	hash->Add(ST_MAP_AMBIENT, ID_MAP_AMBIENT);
	hash->Add(ST_MAP_DIFFUSE, ID_MAP_DIFFUSE);
	hash->Add(ST_MAP_SPECULAR, ID_MAP_SPECULAR);
	hash->Add(ST_MAP_SHINE, ID_MAP_SHINE);
	hash->Add(ST_MAP_SHINESTRENGTH, ID_MAP_SHINESTRENGTH);
	hash->Add(ST_MAP_SELFILLUM, ID_MAP_SELFILLUM);
	hash->Add(ST_MAP_OPACITY, ID_MAP_OPACITY);
	hash->Add(ST_MAP_FILTERCOLOR, ID_MAP_FILTERCOLOR);
	hash->Add(ST_MAP_BUMP, ID_MAP_BUMP);
	hash->Add(ST_MAP_REFLECT, ID_MAP_REFLECT);
	hash->Add(ST_MAP_REFRACT, ID_MAP_REFRACT);
	hash->Add(ST_TM_ROW0, ID_TM_ROW0);
	hash->Add(ST_TM_ROW1, ID_TM_ROW1);
	hash->Add(ST_TM_ROW2, ID_TM_ROW2);
	hash->Add(ST_TM_ROW3, ID_TM_ROW3);
	hash->Add(ST_TM_POS, ID_TM_POS);
	hash->Add(ST_TM_ROTAXIS, ID_TM_ROTAXIS);
	hash->Add(ST_TM_ROTANGLE, ID_TM_ROTANGLE);
	hash->Add(ST_TM_SCALE, ID_TM_SCALE);
	hash->Add(ST_TM_SCALEAXIS, ID_TM_SCALEAXIS);
	hash->Add(ST_TM_SCALEAXISANG, ID_TM_SCALEAXISANG);
	hash->Add(ST_TM_ANIMATION, ID_TM_ANIMATION);
	hash->Add(ST_INHERIT_POS, ID_INHERIT_POS);
	hash->Add(ST_INHERIT_ROT, ID_INHERIT_ROT);
	hash->Add(ST_INHERIT_SCL, ID_INHERIT_SCL);
	hash->Add(ST_FILENAME, ID_FILENAME);
	hash->Add(ST_FIRSTFRAME, ID_FIRSTFRAME);
	hash->Add(ST_LASTFRAME, ID_LASTFRAME);
	hash->Add(ST_FRAMESPEED, ID_FRAMESPEED);
	hash->Add(ST_TICKSPERFRAME, ID_TICKSPERFRAME);
	hash->Add(ST_ENVMAP, ID_ENVMAP);
	hash->Add(ST_STATICBGCOLOR, ID_STATICBGCOLOR);
	hash->Add(ST_ANIMBGCOLOR, ID_ANIMBGCOLOR);
	hash->Add(ST_STATICAMBIENT, ID_STATICAMBIENT);
	hash->Add(ST_ANIMAMBIENT, ID_ANIMAMBIENT);
	hash->Add(ST_VISIBILITY_TRACK, ID_VISIBILITY_TRACK);
	hash->Add(ST_TIMEVALUE, ID_TIMEVALUE);
	hash->Add(ST_COMMENT, ID_COMMENT);
	hash->Add(ST_FILEID, ID_FILEID);
	hash->Add(ST_BOUNDINGBOX_MIN, ID_BOUNDINGBOX_MIN);
	hash->Add(ST_BOUNDINGBOX_MAX, ID_BOUNDINGBOX_MAX);
}

void LoaderASE::SkipSpace()
{
	while (currChar <= ' ' && currChar != '\0')
	{
		if (currChar == '\n')
			nLine++;
		NextChar();
	}
}

void LoaderASE::NextToken()
{
	SkipSpace();

	lenToken = 0;
	if (currChar >= 'A' && currChar <= 'Z' || currChar >= 'a' && currChar <= 'z' || currChar == '*' || currChar == '_')
	{
		while ((currChar >= '0' && currChar <= '9' || currChar >= 'A' && currChar <= 'Z' || currChar >= 'a' && currChar <= 'z' || currChar == '*' || currChar == '_') && lenToken < sizeof(strToken) - 1)
		{
			strToken[lenToken++] = currChar;
			NextChar();
		}
		strToken[lenToken] = '\0';
		while ((currChar >= '0' && currChar <= '9' || currChar >= 'A' && currChar <= 'Z' || currChar >= 'a' && currChar <= 'z' || currChar == '*' || currChar == '_'))
			NextChar();

		if (!hash->Find(strToken, &idToken))
			throw L"Unknown token";
	}
	else if (currChar >= '0' && currChar <= '9' || currChar == '-' || currChar == '+')
	{
		bool isNeg = currChar == '-', isFloat = false;
		if (currChar == '-' || currChar == '+')
		{
			NextChar();
			SkipSpace();
		}

		while (currChar >= '0' && currChar <= '9' && lenToken < sizeof(strToken) - 1)
		{
			strToken[lenToken++] = currChar;
			NextChar();
		}
		if (currChar == '.')
		{
			isFloat = true;
			strToken[lenToken++] = currChar;
			NextChar();
			while (currChar >= '0' && currChar <= '9' && lenToken < sizeof(strToken) - 1)
			{
				strToken[lenToken++] = currChar;
				NextChar();
			}
		}
		strToken[lenToken] = '\0';

		if (isFloat)
		{
			num = (REAL)atof(strToken);
			if (isNeg)
				num = - num;
			inum = -1;
		}
		else
		{
			inum = atoi(strToken);
			if (isNeg)
				inum = - inum;
			num = (REAL)inum;
		}
		idToken = ID_NUM;
	}
	else if (currChar == '"')
	{
		NextChar();
		while (currChar != '"' && currChar != '\0' && lenToken < sizeof(strToken) - 1)
		{
			strToken[lenToken++] = currChar;
			NextChar();
		}
		while (currChar != '"' && currChar != '\0')
			NextChar();
		strToken[lenToken] = '\0';
		NextChar();
		idToken = ID_STRING;
	}
	else if (currChar == '{')
	{
		idToken = ID_LCURVE;
		NextChar();
	}
	else if (currChar == '}')
	{
		idToken = ID_RCURVE;
		NextChar();
	}
	else if (currChar == ':')
	{
		idToken = ID_COLON;
		NextChar();
	}
	else if (currChar == ',')
	{
		idToken = ID_SEMI;
		NextChar();
	}
	else
		idToken = ID_NONE;
}

void LoaderASE::Check(int id)
{
	if (id != idToken)
	{
		wchar msg[256];
		wsprintf(msg, L"Check token error (must be %d)", id);
		throw msg;
	}
	NextToken();
}

REAL LoaderASE::GetNumber()
{
	REAL x = num;
	Check(ID_NUM);
	return x;
}

int LoaderASE::GetINumber()
{
	int x = inum;
	Check(ID_NUM);
	return x;
}

Vector LoaderASE::GetVector()
{
	REAL x = GetNumber();
	REAL y = GetNumber();
	REAL z = GetNumber();
	return Vector(x, y, z);
}

Quat LoaderASE::GetQuat()
{
	REAL x = GetNumber();
	REAL y = GetNumber();
	REAL z = GetNumber();
	REAL angle = GetNumber();
	Quat q;
	MakeQuat(&q, x, y, z, angle);
	return q;
}

Color LoaderASE::GetColor()
{
	REAL r = GetNumber();
	REAL g = GetNumber();
	REAL b = GetNumber();
	return Color(r, g, b);
}

void LoaderASE::SkipBlock()
{
	Check(ID_LCURVE);
	while (idToken != ID_RCURVE && idToken != ID_NONE)
	{
		switch (idToken)
		{
		case ID_LCURVE:
			SkipBlock();
			break;
		default:
			NextToken();
		}
	}
	Check(ID_RCURVE);
}

void LoaderASE::Parse()
{
	while (idToken)
	{
		switch (idToken)
		{
		case ID_FILEID:
			NextToken();
			if (idToken == ID_NUM)
				if (inum != 200)
					throw L"Unknown version";
			Check(ID_NUM);
			break;
		case ID_COMMENT:
			NextToken();
			Check(ID_STRING);
			break;
		case ID_SCENE:
			NextToken();
			ParseScene();
			break;
		case ID_MATERIAL_LIST:
			NextToken();
			ParseMaterialList();
			break;

		case ID_GROUP:
			NextToken();
			ParseGroup(&scene->root);
			break;
		case ID_GEOMETRY:
			NextToken();
			ParseGeometry(&scene->root);
			break;
		case ID_LIGHT:
			NextToken();
			ParseLight(&scene->root);
			break;
		case ID_CAMERA:
			NextToken();
			ParseCamera(&scene->root);
			break;

		case ID_LCURVE:
			SkipBlock();
			break;
		default:
			NextToken();
		}
	}
}

void LoaderASE::ParseScene()
{
	Check(ID_LCURVE);
	while (idToken != ID_RCURVE && idToken != ID_NONE)
	{
		switch (idToken)
		{
		case ID_FIRSTFRAME:
			NextToken();
			firstFrame = GetINumber();
			break;
		case ID_LASTFRAME:
			NextToken();
			lastFrame = GetINumber();
			break;
		case ID_TICKSPERFRAME:
			NextToken();
			ticksPerFrame = GetINumber();
			break;

		case ID_STATICBGCOLOR:
			NextToken();
			scene->backgroundStatic = GetColor();
			break;
		case ID_STATICAMBIENT:
			NextToken();
			scene->ambientStatic = GetColor();
			break;
		case ID_LCURVE:
			SkipBlock();
			break;
		default:
			NextToken();
		}
	}
	Check(ID_RCURVE);

	scene->SetTotalFrame(lastFrame - firstFrame + 1);
}

void LoaderASE::ParseMaterialList()
{
	Check(ID_LCURVE);
	int materialCount = 0;
	while (idToken != ID_RCURVE && idToken != ID_NONE)
	{
		switch (idToken)
		{
		case ID_MATERIAL_COUNT:
			NextToken();
			materialCount = GetINumber();
			break;
		case ID_MATERIAL:
			NextToken();
			ParseMaterial(GetINumber());
			break;
		case ID_LCURVE:
			SkipBlock();
			break;
		default:
			NextToken();
		}
	}
	Check(ID_RCURVE);
}

void LoaderASE::ParseMaterial(int index)
{
	Material *material = new Material();
	scene->materials.Add(material);

	Check(ID_LCURVE);
	while (idToken != ID_RCURVE && idToken != ID_NONE)
	{
		switch (idToken)
		{
		case ID_MATNAME:
		{
			NextToken();
			material->name = strToken;
			Check(ID_STRING);
			break;
		}

		case ID_AMBIENT:
			NextToken();
			material->ambient = GetColor();
			break;
		case ID_DIFFUSE:
			NextToken();
			material->diffuse = GetColor();
			break;
		case ID_SPECULAR:
			NextToken();
			material->specular = GetColor();
			break;
		case ID_SHINE_STRENGTH:
			NextToken();
			material->specularAmount = GetNumber();
			break;
		case ID_SHINE:
			NextToken();
			material->specularLevel = GetNumber();
			break;
		case ID_TRANSPARENCY:
		{
			NextToken();
			REAL opacity = 1 - GetNumber();
			material->opacity = Color(opacity, opacity, opacity);
			break;
		}

		case ID_MAP_AMBIENT:
			NextToken();
			ParseTexture(&material->ambientMap);
			break;
		case ID_MAP_DIFFUSE:
			NextToken();
			ParseTexture(&material->diffuseMap);
			break;
		case ID_MAP_SPECULAR:
			NextToken();
			ParseTexture(&material->specularMap);
			break;
		case ID_MAP_OPACITY:
			NextToken();
			ParseTexture(&material->opacityMap);
			break;
		case ID_MAP_REFLECT:
			NextToken();
			ParseTexture(&material->reflectionMap);
			break;
		case ID_MAP_BUMP:
			NextToken();
			ParseTexture(&material->bumpMap);
			break;

		case ID_LCURVE:
			SkipBlock();
			break;
		default:
			NextToken();
		}
	}
	Check(ID_RCURVE);
}

void LoaderASE::ParseTexture(Texture **tex)
{
	TextureType type = ttNone;
	Texture *texture = NULL;

	Check(ID_LCURVE);
	while (idToken != ID_RCURVE && idToken != ID_NONE)
	{
		switch (idToken)
		{
		case ID_TEXCLASS:
			NextToken();
			if (strcmp(strToken, "Bitmap") == 0)
				type = ttBitmap;
			else if (strcmp(strToken, "Checker") == 0)
			{
				texture = new TextureChessboard();
				type = ttChessboard;
			}
			Check(ID_STRING);
			break;

		case ID_BITMAP:
		{
			NextToken();
			String fileName = strToken;
			try
			{
				texture = new TextureBitmap(fileName.GetString());
			}
			catch (wchar *msgEx)
			{
				wchar msg[256];
				wsprintf(msg, 
					L"Error open bitmap file\nFile: \"%s\"\nException: %s\nAt line %d",
					fileName.GetString(), msgEx, nLine);
				Warning(msg);

				delete texture;
				texture = NULL;
			}
			Check(ID_STRING);
			break;
		}

		case ID_U_OFFSET:
			NextToken();
			texture->offset.x = GetNumber();
			break;
		case ID_V_OFFSET:
			NextToken();
			texture->offset.y = GetNumber();
			break;
		case ID_U_TILING:
			NextToken();
			texture->tiling.x = GetNumber();
			break;
		case ID_V_TILING:
			NextToken();
			texture->tiling.y = GetNumber();
			break;

		case ID_LCURVE:
			SkipBlock();
			break;
		default:
			NextToken();
		}
	}
	Check(ID_RCURVE);

	/*if (!texture)
		throw L"Unknown texture type";*/
	*tex = texture;
}

void LoaderASE::ParseNodeTM(Node *node)
{
	Matrix m;
	MakeIdentity(&m);

	Check(ID_LCURVE);
	while (idToken != ID_RCURVE && idToken != ID_NONE)
	{
		switch (idToken)
		{
		case ID_NODE_NAME:
		{
			NextToken();
			String name = strToken;
			Check(ID_STRING);
			// - "Camera01.Target"
			if (name != node->name)
				node = NULL;
			break;
		}
		case ID_TM_ROW0:
			NextToken();
			m._00 = GetNumber();
			m._01 = GetNumber();
			m._02 = GetNumber();
			break;
		case ID_TM_ROW1:
			NextToken();
			m._10 = GetNumber();
			m._11 = GetNumber();
			m._12 = GetNumber();
			break;
		case ID_TM_ROW2:
			NextToken();
			m._20 = GetNumber();
			m._21 = GetNumber();
			m._22 = GetNumber();
			break;
		case ID_TM_ROW3:
			NextToken();
			m._30 = GetNumber();
			m._31 = GetNumber();
			m._32 = GetNumber();
			break;

		case ID_LCURVE:
			SkipBlock();
			break;
		default:
			NextToken();
		}
	}
	Check(ID_RCURVE);

	if (node)
		node->mTransform = m;
}

void LoaderASE::ParseNodeTMAnimation(Node *node)
{	
	Vector *ap_t = new Vector[scene->GetTotalFrame()];		// The translation components
	Quat *ap_q = new Quat[scene->GetTotalFrame()];			// The essential rotation
	Quat *ap_u = new Quat[scene->GetTotalFrame()];			// The stretch rotation. This is the axis system of the scaling application
	Vector *ap_k = new Vector[scene->GetTotalFrame()];		// The stretch factors. These are the scale factors for x, y and z

	//    
	for (int i = 0; i < scene->GetTotalFrame(); i++)
	{
		ap_t[i] = Vector(0.0f, 0.0f, 0.0f);
		ap_q[i] = Quat(0.0f, 0.0f, 0.0f, 1.0f);
		ap_u[i] = Quat(0.0f, 0.0f, 0.0f, 1.0f);
		ap_k[i] = Vector(1.0f, 1.0f, 1.0f);
	}

	Check(ID_LCURVE);
	while (idToken != ID_RCURVE && idToken != ID_NONE)
	{
		switch (idToken)
		{
		case ID_NODE_NAME:
		{
			NextToken();
			String name = strToken;
			Check(ID_STRING);
			// - "Camera01.Target"
			if (name != node->name)
				node = NULL;
			break;
		}

		case ID_POS_TRACK:
			NextToken();
			Check(ID_LCURVE);
			while (idToken != ID_RCURVE && idToken != ID_NONE)
			{
				Check(ID_POS_SAMPLE);
				int frame = TicksToSceneFrame(GetINumber());
				ap_t[frame] = GetVector();
			}
			Check(ID_RCURVE);
			break;

		case ID_ROT_TRACK:
			NextToken();
			Check(ID_LCURVE);
			while (idToken != ID_RCURVE && idToken != ID_NONE)
			{
				Check(ID_ROT_SAMPLE);
				int frame = TicksToSceneFrame(GetINumber());
				ap_q[frame] = GetQuat();
			}
			Check(ID_RCURVE);
			break;

		case ID_SCALE_TRACK:
			NextToken();
			Check(ID_LCURVE);
			while (idToken != ID_RCURVE && idToken != ID_NONE)
			{
				Check(ID_SCALE_SAMPLE);
				int frame = TicksToSceneFrame(GetINumber());
				ap_k[frame] = GetVector();
				ap_u[frame] = GetQuat();
			}
			Check(ID_RCURVE);
			break;

		case ID_LCURVE:
			SkipBlock();
			break;
		default:
			NextToken();
		}
	}
	Check(ID_RCURVE);

	//    ap_t, ap_q, ap_u, ap_k 
	if (node)
	{
		node->mTransformAnimation = new Matrix[scene->GetTotalFrame()];

		Matrix current;
		Matrix stm, srtm, srtminv, rtm, rtmCurrent, ptmCurrent;
		MakeIdentity(&rtmCurrent);

		for (int i = 0; i < scene->GetTotalFrame(); i++)
		{
			MakeTranslate(&ptmCurrent, ap_t[i].x, ap_t[i].y, ap_t[i].z);

			MakeMatrix(&rtm, ap_q[i]);
			MultMatrixRight(&rtmCurrent, rtm);

			MakeMatrix(&srtm, ap_u[i]);
			MakeScale(&stm, ap_k[i].x, ap_k[i].y, ap_k[i].z);
			ReverseMatrix(&srtminv, srtm);

			MakeIdentity(&current);
			MultMatrixRight(&current, srtminv);
			MultMatrixRight(&current, stm);
			MultMatrixRight(&current, srtm);
			MultMatrixRight(&current, rtmCurrent);
			MultMatrixRight(&current, ptmCurrent);

			node->mTransformAnimation[i] = current;
		}
	}

	delete ap_t;
	delete ap_q;
	delete ap_u;
	delete ap_k;
}

void LoaderASE::ParseGroup(Group *parent)
{
	Group *group = new Group(scene, parent);
	group->name = strToken;
	Check(ID_STRING);

	Check(ID_LCURVE);
	while (idToken != ID_RCURVE && idToken != ID_NONE)
	{
		switch (idToken)
		{
		case ID_GROUP:
			NextToken();
			ParseGroup(group);
			break;
		case ID_GEOMETRY:
			NextToken();
			ParseGeometry(group);
			break;
		case ID_LIGHT:
			NextToken();
			ParseLight(group);
			break;
		case ID_CAMERA:
			NextToken();
			ParseCamera(group);
			break;

		case ID_LCURVE:
			SkipBlock();
			break;
		default:
			NextToken();
		}
	}
	Check(ID_RCURVE);
}

void LoaderASE::ParseGeometry(Group *parent)
{
	Mesh *mesh = new Mesh(scene, parent);

	Check(ID_LCURVE);
	while (idToken != ID_RCURVE && idToken != ID_NONE)
	{
		switch (idToken)
		{
		case ID_NODE_NAME:
			NextToken();
			mesh->name = strToken;
			Check(ID_STRING);
			break;
		case ID_NODE_TM:
			NextToken();
			ParseNodeTM(mesh);
			break;
		case ID_TM_ANIMATION:
			NextToken();
			ParseNodeTMAnimation(mesh);
			break;
		case ID_MESH:
			NextToken();
			ParseMesh(mesh);
			break;
		case ID_MATERIAL_REF:
			NextToken();
			mesh->SetMatRef(GetINumber());
			break;
		case ID_LCURVE:
			SkipBlock();
			break;
		default:
			NextToken();
		}
	}
	Check(ID_RCURVE);
}

void LoaderASE::ParseMesh(Mesh *mesh)
{
	int numVertex = 0, numTVertex = 0, numFaces = 0;
	Vector *vertexList = NULL, *textureList = NULL;

	Matrix mTransformReverse;
	ReverseMatrix(&mTransformReverse, mesh->mTransform);

	Check(ID_LCURVE);
	while (idToken != ID_RCURVE && idToken != ID_NONE)
	{
		switch (idToken)
		{
		case ID_MESH_NUMVERTEX:
			NextToken();
			numVertex = GetINumber();
			vertexList = new Vector[numVertex];
			break;
		case ID_MESH_NUMTVERTEX:
			NextToken();
			numTVertex = GetINumber();
			textureList = new Vector[numTVertex];
			break;
		case ID_MESH_VERTEX_LIST:
			NextToken();
			Check(ID_LCURVE);
			while (idToken != ID_RCURVE && idToken != ID_NONE)
			{
				switch (idToken)
				{
				case ID_MESH_VERTEX:
				{
					NextToken();
					int nVert = GetINumber();
					Vector v;
					Transform(&v, GetVector(), mTransformReverse);
					vertexList[nVert] = v;
					break;
				}
				default:
					NextToken();
				}
			}
			Check(ID_RCURVE);
			break;
		case ID_MESH_TVERTLIST:
			NextToken();
			Check(ID_LCURVE);
			while (idToken != ID_RCURVE && idToken != ID_NONE)
			{
				switch (idToken)
				{
				case ID_MESH_TVERT:
				{
					NextToken();
					int nVert = GetINumber();
					textureList[nVert] = GetVector();
					break;
				}
				default:
					NextToken();
				}
			}
			Check(ID_RCURVE);
			break;
		case ID_MESH_NUMFACES:
			NextToken();
			numFaces = GetINumber();
			break;
		case ID_MESH_NUMTVFACES:
			NextToken();
			if (numFaces != GetINumber())
				throw L"Diferent *MESH_NUMFACES and *MESH_NUMTVFACES";
			break;
		case ID_MESH_FACE_LIST:
			NextToken();
			Check(ID_LCURVE);
			while (idToken != ID_RCURVE && idToken != ID_NONE)
			{
				switch (idToken)
				{
				case ID_MESH_FACE:
				{
					Triangle *tri = new Triangle();
					mesh->Add(tri);

					NextToken();
					int nVert;
					int nFace = GetINumber();
					Check(ID_COLON);

					Check(ID_MESH_FACE_A);
					Check(ID_COLON);
					nVert = GetINumber();
					tri->pt0 = vertexList[nVert];

					Check(ID_MESH_FACE_B);
					Check(ID_COLON);
					nVert = GetINumber();
					tri->pt1 = vertexList[nVert];

					Check(ID_MESH_FACE_C);
					Check(ID_COLON);
					nVert = GetINumber();
					tri->pt2 = vertexList[nVert];
					break;
				}
				default:
					NextToken();
				}
			}
			Check(ID_RCURVE);
			break;
		case ID_MESH_TFACELIST:
			NextToken();
			Check(ID_LCURVE);
			while (idToken != ID_RCURVE && idToken != ID_NONE)
			{
				switch (idToken)
				{
				case ID_MESH_TFACE:
				{
					NextToken();
					int nVert;
					int nFace = GetINumber();

					Triangle *tri = mesh->GetTri(nFace);
					
					nVert = GetINumber();
					tri->texture0 = textureList[nVert];
					nVert = GetINumber();
					tri->texture1 = textureList[nVert];
					nVert = GetINumber();
					tri->texture2 = textureList[nVert];
					break;
				}
				default:
					NextToken();
				}
			}
			Check(ID_RCURVE);
			break;
		case ID_MESH_NORMALS:
			NextToken();
			Check(ID_LCURVE);
			while (idToken != ID_RCURVE && idToken != ID_NONE)
			{
				switch (idToken)
				{
				case ID_MESH_FACENORMAL:
				{
					NextToken();
					int nFace = GetINumber();
					Triangle *tri = mesh->GetTri(nFace);

					GetVector();

					Check(ID_MESH_VERTEXNORMAL);
					Check(ID_NUM);
					tri->normal0 = GetVector();

					Check(ID_MESH_VERTEXNORMAL);
					Check(ID_NUM);
					tri->normal1 = GetVector();

					Check(ID_MESH_VERTEXNORMAL);
					Check(ID_NUM);
					tri->normal2 = GetVector();
					break;
				}
				default:
					NextToken();
				}
			}
			Check(ID_RCURVE);
			break;
		case ID_LCURVE:
			SkipBlock();
			break;
		default:
			NextToken();
		}
	}
	Check(ID_RCURVE);

	delete vertexList;
	delete textureList;
}

void LoaderASE::ParseLight(Group *parent)
{
	Light *light = new Light(scene, parent);
	scene->lights.Add(light);

	Check(ID_LCURVE);
	while (idToken != ID_RCURVE && idToken != ID_NONE)
	{
		switch (idToken)
		{
		case ID_NODE_NAME:
			NextToken();
			light->name = strToken;
			Check(ID_STRING);
			break;
		case ID_NODE_TM:
			NextToken();
			ParseNodeTM(light);
			break;
		case ID_TM_ANIMATION:
			NextToken();
			ParseNodeTMAnimation(light);
			break;
		case ID_LIGHT_SETTINGS:
			NextToken();
			ParseLightSettings(light);
			break;
		case ID_LCURVE:
			SkipBlock();
			break;
		default:
			NextToken();
		}
	}
	Check(ID_RCURVE);
}

void LoaderASE::ParseLightSettings(Light *light)
{
	Check(ID_LCURVE);
	while (idToken != ID_RCURVE && idToken != ID_NONE)
	{
		switch (idToken)
		{
		case ID_LIGHT_COLOR:
			NextToken();
			light->color = GetColor();
			break;
		case ID_LIGHT_INTENS:
			NextToken();
			light->intens = GetNumber();
			break;
		case ID_LCURVE:
			SkipBlock();
			break;
		default:
			NextToken();
		}
	}
	Check(ID_RCURVE);
}

void LoaderASE::ParseCamera(Group *parent)
{
	Camera *camera = new Camera(scene, parent);
	scene->cameras.Add(camera);

	Check(ID_LCURVE);
	while (idToken != ID_RCURVE && idToken != ID_NONE)
	{
		switch (idToken)
		{
		case ID_NODE_NAME:
			NextToken();
			camera->name = strToken;
			Check(ID_STRING);
			break;
		case ID_NODE_TM:
			NextToken();
			ParseNodeTM(camera);
			break;
		case ID_TM_ANIMATION:
			NextToken();
			ParseNodeTMAnimation(camera);
			break;
		case ID_LCURVE:
			SkipBlock();
			break;
		default:
			NextToken();
		}
	}
	Check(ID_RCURVE);
}
