/******************************************************************************
*                                                                             *
*                                                    *
*                                                     *
*                                                                             *
*   15  2004 .                                                        *
*   (C)                                                         *
*   e-mail       pavelkok@mail.ru                                             *
*   home page    http://kps.nm.ru                                             *
*                                                                             *
*                                                                             *
*   main.cpp                                                                  *
*      Windows 2000/XP                                       *
*                                                                             *
*                                                                             *
*******************************************************************************/


#include "types.h"
#include "scene.h"
#include "impt_ase.h"

#include "dlg_materials.h"
#include "resource.h"


int nWidth = 800;
int nHeight = 600;


HWND hwnd;
HINSTANCE hInstance;
int iCamera = 0;						// 0 -  , 1..9 

REAL AngleView = PI/2;
REAL AngleX = PI/12;
REAL AngleY = 0;
REAL Pos = - 100.0f;

REAL dAngleView = PI/72;
REAL dAngleX = PI/36;
REAL dAngleY = PI/36;
REAL dPos = 5.0f;

bool isDrawTracing = false;				//   
bool needCancel = false;				//   ESC
SceneState sceneState = ssIdle;
bool isRedrawNow = false, isAnimationNow = false;

Scene *scene;
uint ticksRenderStart;


const wchar stClassName[] = L"RayTracing";

const wchar stAppName[] = L"Ray Tracing";
const wchar stAppNameFrame[] = L"Ray Tracing (frame %d of %d)";
const wchar stAppRender[] = L"Render (frame %d of %d, %d%% done, %0.2f sec)";
const wchar stAppRenderDone[] = L"Ray Tracing (Rendering %0.2f sec)";

const wchar stAbout[] =
	L"Ray Tracing\n"
	L"Pavel Kokolemin\n\n"
	L"Using: raytracing.exe *.ase\n\n"
	L"Keys: \n"
	L"F1 - About\n"
	L"F2 - Scene statistics\n"
	L"F5 - Save window shot\n"
	L"F9 - Material editor\n"
	L"Tab/Shift+Tab - next/prior animation frame\n"
	L"Enter - toggle draw and render mode\n"
	L"Left, Right, Up, Down - rotate view\n"
	L"PageUp, PageDown - move view\n"
	L"'-', '=' - zoom plus/minus\n"
	L"Ctrl + 0 - user view\n"
	L"Ctrl + 1..9 - cameraN view\n"
	L"1..8 - change cube count\n"
	L"Esc - Exit\n\n"
	L"Release info\n"
	L"sizeof(REAL)\t\t\t%d\n"
	L"MAX_CUBE_COUNT\t\t\t%d\n"
	L"TRACING_LEVEL\t\t\t%d\n"
	L"ADAPTIVE_ANTIALIASING_DIFF\t%0.2f\n";

const wchar stSceneInfo[] = 
	L"Triangles\t\t\t%d\n"
	L"Lights\t\t\t%d\n"
	L"Cameras\t\t\t%d\n"
	L"Materials\t\t\t%d\n\n"
	L"Cubes Per Line\t\t%d\n"
	L"Cubes Total\t\t%d\n"
	L"Cubes Used\t\t%d\n\n"
	L"Tri Total\t\t\t%d\n"
	L"Tri Peak\t\t\t%d\n"
	L"Tri Average\t\t%0.4f\n";


const int iRedrawRenderLines = 8;

const int iTimerID = 1;
const int iTimerRate = 1000;


LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
BOOL Cls_OnCreate(HWND hwnd, LPCREATESTRUCT lpCreateStruct);
void Cls_OnDestroy(HWND hwnd);
void Cls_OnTimer(HWND hwnd, UINT id);
void Cls_OnPaint(HWND hwnd);
void Cls_OnLButtonDown(HWND hwnd, BOOL fDoubleClick, int x, int y, UINT keyFlags);
void Cls_OnLButtonUp(HWND hwnd, int x, int y, UINT keyFlags);
void Cls_OnMouseMove(HWND hwnd, int x, int y, UINT keyFlags);
void Cls_OnKey(HWND hwnd, UINT vk, BOOL fDown, int cRepeat, UINT flags);



//   
void ProcessMessages()
{
	MSG msg;
	while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
}

void ShutdownWindows()
{
	HANDLE hToken;
	TOKEN_PRIVILEGES tkp;

	// Get a token for this process. 
	if (!OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken))
		throw L"OpenProcessToken";

	// Get the LUID for the shutdown privilege. 
	LookupPrivilegeValue(NULL, SE_SHUTDOWN_NAME, &tkp.Privileges[0].Luid); 

	tkp.PrivilegeCount = 1;  // one privilege to set
	tkp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;

	// Get the shutdown privilege for this process.
	AdjustTokenPrivileges(hToken, FALSE, &tkp, 0, (PTOKEN_PRIVILEGES)NULL, 0);

	// Cannot test the return value of AdjustTokenPrivileges.
	if (GetLastError() != ERROR_SUCCESS)
		throw L"AdjustTokenPrivileges";

	// Shut down the system and force all applications to close. 
 
	if (!ExitWindowsEx(EWX_POWEROFF /*| EWX_FORCE*/, 0))
		throw L"ExitWindowsEx"; 
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	INITCOMMONCONTROLSEX icce;
	icce.dwSize = sizeof(icce);
	icce.dwICC = ICC_USEREX_CLASSES;
	InitCommonControlsEx(&icce);

	WNDCLASSEX wndClass;
	MSG msg;
	
	::hInstance = hInstance;

	//  
	wndClass.cbSize = sizeof(wndClass);
	wndClass.style = CS_HREDRAW | CS_VREDRAW;
	wndClass.lpfnWndProc = WndProc;
	wndClass.cbClsExtra = 0;
	wndClass.cbWndExtra = 0;
	wndClass.hInstance = hInstance;
	wndClass.hIcon = LoadIcon(0, IDI_APPLICATION);
	wndClass.hIconSm = LoadIcon(0, IDI_APPLICATION);
	wndClass.hCursor = LoadCursor(0, IDC_ARROW);
	wndClass.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
	wndClass.lpszMenuName = NULL;
	wndClass.lpszClassName = stClassName;
	
	RegisterClassEx(&wndClass);

	if (nWidth == 0)
		nWidth = GetSystemMetrics(SM_CXSCREEN);
	if (nHeight == 0)
		nHeight = GetSystemMetrics(SM_CYSCREEN);

	//  
	hwnd = CreateWindowEx(
		0, //  
		stClassName, //  
		stAppName, //  
		WS_POPUP | WS_CAPTION | WS_MINIMIZEBOX | WS_SYSMENU , //  
		0, 0, // 
		nWidth, // 
		nHeight, // 
		0,  //   
		0,  //  
		hInstance,  //  
		0);  //  

	RECT cr, wr;
	GetClientRect(hwnd, &cr);
	GetWindowRect(hwnd, &wr);
	SetWindowPos(hwnd, HWND_TOP, 
		(GetSystemMetrics(SM_CXSCREEN) - nWidth)/2, 
		(GetSystemMetrics(SM_CYSCREEN) - nHeight)/2, 
		nWidth + wr.right - wr.left - cr.right, 
		nHeight + wr.bottom - wr.top - cr.bottom, 
		SWP_SHOWWINDOW);
    
	//   
	while (GetMessage(&msg, NULL, 0, 0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	
	return (int)msg.wParam;
}

//  
LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	try
	{
		switch (msg)
		{
		case WM_CREATE:
			return Cls_OnCreate(hwnd, LPCREATESTRUCT(lParam))? 0: -1;
		case WM_DESTROY:
			Cls_OnDestroy(hwnd); 
			return 0;
		case WM_TIMER:
			Cls_OnTimer(hwnd, (UINT)wParam); 
			return 0;
		case WM_PAINT:
			Cls_OnPaint(hwnd); 
			return 0;
		case WM_LBUTTONDOWN:
			Cls_OnLButtonDown(hwnd, FALSE, (int)(short)LOWORD(lParam), (int)(short)HIWORD(lParam), (UINT)(wParam));
			return 0;
		case WM_MOUSEMOVE:
			Cls_OnMouseMove(hwnd, (int)(short)LOWORD(lParam), (int)(short)HIWORD(lParam), (UINT)(wParam));
			return 0;
		case WM_LBUTTONUP:
			Cls_OnLButtonUp(hwnd, (int)(short)LOWORD(lParam), (int)(short)HIWORD(lParam), (UINT)(wParam));
			return 0;
		case WM_KEYDOWN:
			Cls_OnKey(hwnd, (UINT)(wParam), TRUE, (int)(short)LOWORD(lParam), (UINT)HIWORD(lParam));
			return 0;
		case WM_KEYUP:
			Cls_OnKey(hwnd, (UINT)(wParam), FALSE, (int)(short)LOWORD(lParam), (UINT)HIWORD(lParam));
			return 0;
		default:
			return DefWindowProc(hwnd, msg, wParam, lParam);
		}
	}
	catch (wchar* msg)
	{
		MessageBox(hwnd, msg, L"Error!", MB_ICONERROR);
		return 0;
	}
}


void CallbackFunc(SceneState state, int done, int total, bool *cancel)
{
	sceneState = state;
	ProcessMessages();
	if ((cancel != NULL) && needCancel)
		*cancel = true;

	switch (state)
	{
	case ssRenderStart:
		ticksRenderStart = GetTickCount();
		break;
	case ssPrepareTriList:
		SetWindowText(hwnd, L"Prepare TriList");
		break;
	case ssCreateBoundingBox:
		SetWindowText(hwnd, L"Create Bounding Box");
		break;
	case ssCreateCubes:
		SetWindowText(hwnd, L"Create Cubes");
		break;
	case ssRender:
		if (done%iRedrawRenderLines == 0)
		{
			float time = (GetTickCount() - ticksRenderStart)/1000.0f;
			int percent = done*100/total;

			wchar msg[256];
			swprintf(msg, stAppRender, scene->GetCurrentFrame(), scene->GetTotalFrame(), percent, time);
			SetWindowText(hwnd, msg);

			HDC hdc = GetDC(hwnd);
			BitBlt(hdc, 
				0, done - iRedrawRenderLines, scene->buffer->GetWidth(), iRedrawRenderLines,
				scene->buffer->GetDC(), 0, done - iRedrawRenderLines, SRCCOPY);
			ReleaseDC(hwnd, hdc);
		}
		break;
	case ssRenderDone:
	{
		float time = (GetTickCount() - ticksRenderStart)/1000.0f;
		wchar msg[256];
		swprintf(msg, stAppRenderDone, time);
		SetWindowText(hwnd, msg);
		sceneState = ssIdle;
		break;
	}
	}
}

void RedrawBuffer()
{
	needCancel = false;
	if (iCamera == 0)
	{
		MakeIdentity(&scene->mView);
		RotateX(&scene->mView, - PI/2);
		RotateY(&scene->mView, AngleY);
		RotateX(&scene->mView, AngleX);
		Translate(&scene->mView, 0, 0, Pos);
	}
	else
	{
		Camera *camera = scene->cameras[iCamera - 1];
		if (camera->mTransformAnimation)
			ReverseMatrix(&scene->mView, camera->mTransformAnimation[scene->GetCurrentFrame()]);
		else
			ReverseMatrix(&scene->mView, camera->mTransform);
	}

	isRedrawNow = true;
	if (!isDrawTracing)
		scene->Render();
	else
		scene->RenderTracing();
	isRedrawNow = false;

	InvalidateRect(hwnd, NULL, FALSE);
	UpdateWindow(hwnd);
}


void SaveShot()
{
	SYSTEMTIME time;
	GetLocalTime(&time);
	wchar_t fileName[MAX_PATH];
	wsprintf(fileName, L"shots\\%04d%02d%02d-%02d%02d%02d.png", 
		time.wYear, time.wMonth, time.wDay,
		time.wHour, time.wMinute, time.wSecond);

	CreateDirectory(L"shots", NULL);

	SetWindowText(hwnd, L"Save to file...");
	scene->buffer->SaveToFile(fileName);
	SetWindowText(hwnd, stAppName);
}


void SaveAnimation()
{
	SYSTEMTIME time;
	GetLocalTime(&time);
	wchar_t fileMask[MAX_PATH];
	wsprintf(fileMask, L"animation\\%04d%02d%02d-%02d%02d%02d-%%04d.bmp", 
		time.wYear, time.wMonth, time.wDay,
		time.wHour, time.wMinute, time.wSecond);

	wchar_t fileName[MAX_PATH];

	CreateDirectory(L"animation", NULL);

	needCancel = false;
	isAnimationNow = true;
	for (int frame = scene->GetCurrentFrame(); frame < scene->GetTotalFrame(); frame++)
	{
		scene->SetCurrentFrame(frame);
		RedrawBuffer();

		wsprintf(fileName, fileMask, frame);
		scene->buffer->SaveToFile(fileName);

		ProcessMessages();
		if (needCancel)
			break;
	}
	isAnimationNow = false;

	// ,     ,  
	if (isDrawTracing && !needCancel)
		ShutdownWindows();
}


BOOL Cls_OnCreate(HWND hwnd, LPCREATESTRUCT lpCreateStruct)
{
	srand(GetTickCount());

	//SetTimer(hwnd, iTimerID, iTimerRate, NULL);

	scene = new Scene(nWidth, nHeight, AngleView);
	scene->callbackFunc = &CallbackFunc;

	wchar *cmdLine = GetCommandLine();
	int nArgs;
	wchar **cmds = CommandLineToArgvW(cmdLine, &nArgs);
	if (nArgs == 2)
		LoaderASE(cmds[1], scene);

	RedrawBuffer();

	return TRUE;
}

void Cls_OnDestroy(HWND hwnd)
{
	//KillTimer(hwnd, iTimerID);

	delete scene;

	PostQuitMessage(0);
}

void Cls_OnTimer(HWND hwnd, UINT id)
{	
}

void Cls_OnPaint(HWND hwnd)
{
	PAINTSTRUCT ps;
	HDC hdc = BeginPaint(hwnd, &ps);

	BitBlt(hdc, 0, 0, scene->GetWidth(), scene->GetHeight(), scene->GetDC(), 0, 0, SRCCOPY);

	EndPaint(hwnd, &ps);
}

void Cls_OnLButtonDown(HWND hwnd, BOOL fDoubleClick, int x, int y, UINT keyFlags)
{
	wchar msg[256];
	swprintf(msg, L"x: %d, y: %d", x, y);
	SetWindowText(hwnd, msg);
}

void Cls_OnLButtonUp(HWND hwnd, int x, int y, UINT keyFlags)
{
}

void Cls_OnMouseMove(HWND hwnd, int x, int y, UINT keyFlags)
{
}

void Cls_OnKey(HWND hwnd, UINT vk, BOOL fDown, int cRepeat, UINT flags)
{
	if (!fDown)
		return;

	if (!isRedrawNow && !isAnimationNow)
	{
		switch (vk)
		{
		case VK_ESCAPE:
			DestroyWindow(hwnd);
			break;

		case VK_F1:
		{
			wchar buf[1024];
			swprintf(buf, stAbout, sizeof(REAL), MAX_CUBE_COUNT, TRACING_LEVEL, ADAPTIVE_ANTIALIASING_DIFF);
			MessageBox(hwnd, buf, L"About...", MB_ICONINFORMATION);
			break;
		}

		case VK_F2:
		{
			SceneInfo info;
			scene->GetSceneInfo(&info);

			wchar buf[1024];
			swprintf(buf, stSceneInfo, 
				info.trianglesCount, info.lightsCount, info.camerasCount, info.materialsCount, 
				info.cubesPerLine, info.cubesTotal, info.cubesUsed, 
				info.triTotal, info.triPeak, info.triAverage);
			MessageBox(hwnd, buf, L"Scene info...", MB_ICONINFORMATION);
			break;
		}

		case VK_F5:
			if (GetKeyState(VK_CONTROL) & 0x8000)
				SaveAnimation();
			else
				SaveShot();
			break;

		case VK_F9:
			if (DialogBoxParam(hInstance, MAKEINTRESOURCE(IDD_MATERIAL), hwnd, &MaterialDialogProc, 0))
				RedrawBuffer();
			break;

		case VK_RETURN:
			isDrawTracing = !isDrawTracing;
			if (!isDrawTracing)
				SetWindowText(hwnd, stAppName);
			RedrawBuffer();
			break;

		case VK_LEFT:
		case VK_RIGHT:
		case VK_UP:
		case VK_DOWN:
		case VK_PRIOR:
		case VK_NEXT:
		{
			if (iCamera != 0)
				break;
			switch (vk)
			{
				case VK_LEFT:	AngleY -= dAngleY; break;
				case VK_RIGHT:	AngleY += dAngleY; break;
				case VK_UP:		AngleX -= dAngleX; break;
				case VK_DOWN:	AngleX += dAngleX; break;
				case VK_PRIOR:	Pos += dPos; break;
				case VK_NEXT:	Pos -= dPos; break;
			}			
			RedrawBuffer();
			break;
		}

		case VK_TAB:
		{
			int frame = scene->GetCurrentFrame();
			if (GetKeyState(VK_SHIFT) & 0x8000)
			{
				frame--;
				if (frame < 0)
					frame = scene->GetTotalFrame() - 1;
			}
			else
			{
				frame++;
				if (frame >= scene->GetTotalFrame())
					frame = 0;
			}
			scene->SetCurrentFrame(frame);
			RedrawBuffer();

			wchar msg[256];
			wsprintf(msg, stAppNameFrame, scene->GetCurrentFrame(), scene->GetTotalFrame());
			SetWindowText(hwnd, msg);
			break;
		}

		case VK_SPACE:
			scene->Test();
			RedrawBuffer();
			break;

		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
		{
			int index = vk - '0';

			//  
			if (GetKeyState(VK_CONTROL) & 0x8000)
			{
				if (index - 1 >= scene->cameras.GetCount())
					throw L"No camera with this index";

				if (index != iCamera)
				{
					iCamera = index;
					RedrawBuffer();
				}
			}			
			else //  - 
			{
				if (index >= 1 && index <= 8)
				{
					int count = (int)POW(2, (REAL)index);
					scene->SetCubeCount(count);

					wchar msg[256];
					wsprintf(msg, L"Cubes per line set to %d", count);
					SetWindowText(hwnd, msg);
				}
			}
			break;
		}

		case VK_OEM_MINUS:
			scene->SetAngleView(scene->GetAngleView() - dAngleView);
			RedrawBuffer();
			break;
		case VK_OEM_PLUS:
			scene->SetAngleView(scene->GetAngleView() + dAngleView);
			RedrawBuffer();
			break;
	}
	}
	else
	{
		switch (vk)
		{
		case VK_ESCAPE:
			needCancel = true;
			break;
		}
	}
}
