/******************************************************************************
*                                                                             *
*                                                    *
*                                                     *
*                                                                             *
*   15  2004 .                                                        *
*   (C)                                                         *
*   e-mail       pavelkok@mail.ru                                             *
*   home page    http://kps.nm.ru                                             *
*                                                                             *
*                                                                             *
*   material.h                                                                *
*                                     *
*                                                                             *
*******************************************************************************/


#ifndef MATERIAL_H
#define MATERIAL_H

#include "types.h"
#include "bitmap.h"


enum TextureType
{
	ttNone, ttBitmap, ttChessboard, ttPerlinNoise
};


class Texture
{
private:
	TextureType type;
protected:
	virtual void EvaluteColor(const Vector &v, Color *color) = NULL;
	virtual void EvaluteBump(const Vector &v, Vector *bump);
	Texture(Texture* tex);
	Texture(TextureType type);
public:
	String name;
	Vector offset;
	Vector tiling;
	bool invert;
	virtual ~Texture();
	Texture *Copy();
	TextureType GetType() { return type; }
	void GetColor(const Vector &v, Color *color);
	void GetBump(const Vector &v, Vector *bump);
};


class TextureBitmap: public Texture
{
private:
	Bitmap bitmap;
protected:
	REAL Interpolate(REAL v1, REAL v2, REAL t);
	void EvaluteColor(const Vector &v, Color *color);
	void EvaluteBump(const Vector &v, Vector *bump);
public:
	TextureBitmap(wchar *fileName);
	TextureBitmap(TextureBitmap* tex);
	~TextureBitmap();
	String fileName;
	Bitmap *GetBitmap() { return &bitmap; }
	int GetWidth() { return bitmap.GetWidth(); }
	int GetHeight() { return bitmap.GetHeight(); }
};


class TextureChessboard: public Texture
{
private:
protected:
	void EvaluteColor(const Vector &v, Color *color);
public:
	Color evenColor;
	Color oddColor;
	TextureChessboard();
	TextureChessboard(TextureChessboard* tex);
};


class TexturePerlinNoise: public Texture
{
private:
	REAL Noise(int x, int y, int z, int t);
	REAL SmoothedNoise(int x, int y, int z, int t);
	REAL Interpolate(REAL v1, REAL v2, REAL t);
	REAL FinalNoise(REAL x, REAL y, REAL z, int t);
protected:
	void EvaluteColor(const Vector &v, Color *color);
public:
	Color baseColor;
	Color noiseColor;
	REAL amplitude, frequency, persistence;
	int octaves;
	TexturePerlinNoise(REAL amplitude, REAL frequency, REAL persistence, int octaves);
	TexturePerlinNoise(TexturePerlinNoise* tex);
};


class Material
{
private:
public:
	String name;
	Color ambient;
	Color diffuse;
	Color specular;
	Color opacity;
	Color reflection;
	Texture *ambientMap;
	Texture *diffuseMap;
	Texture *specularMap;
	Texture *opacityMap;
	Texture *reflectionMap;
	Texture *bumpMap;
	REAL specularAmount;					// MATERIAL_SHINESTRENGTH
	REAL specularLevel;						// MATERIAL_SHINE
	REAL refractionLevel;
	REAL bumpAmount;
	Material();
	~Material();
	virtual Material *Copy();
};


const int MatRefDefault = -1;
extern Material MaterialDefault;


#endif
