/******************************************************************************
*                                                                             *
*                                                    *
*                                                     *
*                                                                             *
*   15  2004 .                                                        *
*   (C)                                                         *
*   e-mail       pavelkok@mail.ru                                             *
*   home page    http://kps.nm.ru                                             *
*                                                                             *
*                                                                             *
*   nodes.h                                                                   *
*                                                              *
*                                                                             *
*******************************************************************************/


#ifndef NODES_H
#define NODES_H

#include "types.h"
#include "triangle.h"
#include "material.h"


class Node
{
	friend Scene;
	friend Group;
protected:
	Scene *scene;
	Group *parent;
	Matrix mTransformWorld;				//   ( mTransform)
public:
	String name;
	Matrix mTransform;
	Matrix *mTransformAnimation;
	Node(Scene *scene, Group *parent);
	virtual ~Node();
	virtual void Prepare(const Matrix &m, bool render, TriangleArray *list);
	REAL GetX() { return mTransformWorld._30; }
	REAL GetY() { return mTransformWorld._31; }
	REAL GetZ() { return mTransformWorld._32; }
	Vector GetPos() { return Vector(GetX(), GetY(), GetZ()); }
};


class Group: public Node
{
	friend Scene;
private:
	NodeArray subNodes;
public:
	Group(Scene *scene, Group *parent);
	~Group();
	void Prepare(const Matrix &m, bool render, TriangleArray *list);
	int GetCount() { return subNodes.GetCount(); }
	Node *GetNode(int index) { return subNodes[index]; }
	void Add(Node *node) { subNodes.Add(node); }
	void Remove(Node *node) { subNodes.Remove(node); }
};


class Mesh: public Node
{
	friend Scene;
private:
	TriangleArray triangles;
	int matRef;
	Material *mat;
public:
	Mesh(Scene *scene, Group *parent);
	void Prepare(const Matrix &m, bool render, TriangleArray *list);

	void Add(Triangle *tri) { triangles.Add(tri); }
	Triangle *GetTri(int index) { return triangles[index]; }
	int GetCount() { return triangles.GetCount(); }
	int GetMatRef() { return matRef; }
	void SetMatRef(int matRef);
	Material *GetMat(Material *mat) { return mat; }
	void SetMat(Material *mat);
};


enum LightType
{
	ltOmni, ltSpot, ltDirectional
};

enum LightDecay
{
	ldNone, ldInverse, ldInverseSq
};

class Light: public Node
{
	friend Scene;
private:
	Vector dirw;						//   . 
	REAL cosHotspot, cosFalloff;
	Mesh view;							//  .    
	Material matView;
public:
	Color color;
	REAL intens;
	LightType type;
	LightDecay decay;
	REAL decayStart;
	REAL coneHotspot, coneFalloff;		//   (Hotspot)   (Falloff)
	Light(Scene *scene, Group *parent);
	void Prepare(const Matrix &m, bool render, TriangleArray *list);
	//    
	void EvaluteDir(const Vector &pt, Vector *dir);
	//   
	void EvaluteLight(const Vector &pt, const Vector &dir, Color *color);
};


class Camera: public Node
{
	friend Scene;
private:
	Mesh view;							//     
	Material matView;
public:
	Camera(Scene *scene, Group *parent);
	void Prepare(const Matrix &m, bool render, TriangleArray *list);
};


#endif
