/******************************************************************************
*                                                                             *
*                                                    *
*                                                     *
*                                                                             *
*   15  2004 .                                                        *
*   (C)                                                         *
*   e-mail       pavelkok@mail.ru                                             *
*   home page    http://kps.nm.ru                                             *
*                                                                             *
*                                                                             *
*   scene.h                                                                   *
*                                                               *
*                                                                             *
*******************************************************************************/


#ifndef SCENE_H
#define SCENE_H

#include <math.h>

#include "types.h"
#include "bitmap.h"
#include "triangle.h"
#include "material.h"
#include "nodes.h"
#include "impt_ase.h"


//     CUBE_COUNT^3  
#define MAX_CUBE_COUNT					256
#define TRACING_LEVEL					3

//   
#define ADAPTIVE_ANTIALIASING_DIFF		0.5f

//    Material::specular  
#define SPECULARLEVEL_MIN				0.5f
#define SPECULARLEVEL_SCALE				100.0f


enum SceneState
{
	ssIdle, ssRenderStart, ssPrepareTriList, ssCreateBoundingBox, ssCreateCubes, ssRender, ssRenderDone
};

typedef void SceneStateCallback(SceneState state, int done, int total, bool *cancel);

struct SceneInfo
{
	int trianglesCount;
	int lightsCount;
	int camerasCount;
	int materialsCount;
	int cubesPerLine;
	int cubesTotal;
	int cubesUsed;
	int triTotal;
	int triPeak;
	REAL triAverage;
};

struct CubeInfo
{
	int ix, iy, iz;
};


class Scene
{
	friend LoaderASE;
private:
	PSimpleTriangleArray cubes[MAX_CUBE_COUNT][MAX_CUBE_COUNT][MAX_CUBE_COUNT];
	int cubeCount;

	Vector minBounding, maxBounding, dBounding, dBoundingReverse;
	int nWidth, nHeight;
	REAL angleView, zProjection, zDrawClipNear;

	Color backgroundStatic, ambientStatic;

	int currentFrame, totalFrame;

	void CreateBuffer();
	void DeleteBuffer();
	void CreateBoundingBox(PTriangleArray list);
	void CreateCubes(PTriangleArray list);
	void DeleteCubes();
	void AddOnceTriToCube(int i, int j, int k, PTriangle ptri);

	REAL GetCubeX(int i) { return minBounding.x + dBounding.x*i; }
	REAL GetCubeY(int i) { return minBounding.y + dBounding.y*i; }
	REAL GetCubeZ(int i) { return minBounding.z + dBounding.z*i; }
	int GetCubeIX(REAL v) { return (int)((v - minBounding.x)*dBoundingReverse.x); }
	int GetCubeIY(REAL v) { return (int)((v - minBounding.y)*dBoundingReverse.y); }
	int GetCubeIZ(REAL v) { return (int)((v - minBounding.z)*dBoundingReverse.z); }

	bool FirstCube(CubeInfo *cube, Vector *p, const Vector &dir);
	bool NextCube(CubeInfo *cube, Vector *p, const Vector &dir);

	bool Intersect(const Vector &o, const Vector &dir, REAL *t, PTriangle *ppTri);

	void RayColor(const Vector &oRay, const Vector &dirRay_, int recurseLevel, Color *outColor);

	void Callback(SceneState state, int done, int total, bool *cancel);
public:
	Group root;
	LightArray lights;
	CameraArray cameras;
	MaterialArray materials;
	Bitmap *buffer;

	Matrix mView, mViewReverse;

	SceneStateCallback *callbackFunc;

	Scene(DWORD nWidth, DWORD nHeight, REAL angleView);
	~Scene();

	void ViewToScreen(const Vector &v, REAL *x, REAL *y);
	void ScreenToView(Vector *res, REAL x, REAL y);
	void WorldToScreen(const Vector &v, REAL *x, REAL *y);
	void ScreenToWorld(Vector *res, REAL x, REAL y);
	void ViewToScreeni(const Vector &v, int *ix, int *iy);
	void ScreenToViewi(Vector *res, int ix, int iy);
	void WorldToScreeni(const Vector &v, int *ix, int *iy);
	void ScreenToWorldi(Vector *res, int ix, int iy);

	void ClearBuffer(COLORREF color);
	void DrawLine(const Vector &p1, const Vector &p2, COLORREF color);
	void DrawVector(const Vector &p, const Vector &dir, COLORREF color);
	void DrawAxis(REAL size);
	void DrawTriangle(const Triangle &t, COLORREF color);
	void DrawCube(int ix, int iy, int iz, COLORREF color);
	void DrawCube(const Vector &pt0, const Vector &pt1, COLORREF color);

	void Test();

	void GetSceneInfo(SceneInfo *info);

	HDC GetDC() { return buffer->GetDC(); }
	int GetWidth() { return buffer->GetWidth(); }
	int GetHeight() { return buffer->GetHeight(); }
	REAL GetAngleView() { return angleView; }
	void SetAngleView(REAL value);
	int GetCubeCount() { return cubeCount; }
	void SetCubeCount(int value);

	// 
	int GetCurrentFrame() { return currentFrame; }
	int GetTotalFrame() { return totalFrame; }
	void SetCurrentFrame(int frame);
	void SetTotalFrame(int total);

	void Render();
	void RenderTracingLine(const Vector &vScreenO, int j, Color *line);
	void RenderTracingLinePlus(const Vector &vScreenO, int j, Color *line);
	void RenderTracingLineAA(const Vector &vScreenO, int j, Color *line, Color *linePrev, Color *lineCurr, Color *lineNext);
	void RenderTracing();
};

#endif
