#include "triangle.h"

Triangle::Triangle()
{
}

Triangle::~Triangle()
{
}

void Triangle::Prepare()
{
	VectorSubtraction(&vw1, ptw1, ptw0);
	VectorSubtraction(&vw2, ptw2, ptw0);

	VectorProduct(&normalw, vw1, vw2);
	normalw.Normalize();
	planeD = ScalarProduct(normalw, ptw0);

	REAL sXY = vw1.x*vw2.y - vw2.x*vw1.y, asXY = FABS(sXY);
	REAL sYZ = vw1.y*vw2.z - vw2.y*vw1.z, asYZ = FABS(sYZ);
	REAL sZX = vw1.z*vw2.x - vw2.z*vw1.x, asZX = FABS(sZX);

	if (asXY > asYZ && asXY > asZX)
	{
		validPlane = axpXY;
		ukx =   vw2.y/sXY;
		uky = - vw2.x/sXY;
		vkx = - vw1.y/sXY;
		vky =   vw1.x/sXY;
	}
	else if (asYZ > asZX)
	{
		validPlane = axpYZ;
		ukx =   vw2.z/sYZ;
		uky = - vw2.y/sYZ;
		vkx = - vw1.z/sYZ;
		vky =   vw1.y/sYZ;
	}
	else if (asZX > 0)
	{
		validPlane = axpZX;
		ukx =   vw2.x/sZX;
		uky = - vw2.z/sZX;
		vkx = - vw1.x/sZX;
		vky =   vw1.z/sZX;
	}
	else
	{
		validPlane = 0;
		ukx = uky = vkx = vky = 0;
	}

	REAL du1 = texture1.x - texture0.x;
	REAL du2 = texture2.x - texture0.x;
	REAL dv1 = texture1.y - texture0.y;
	REAL dv2 = texture2.y - texture0.y;
	REAL d = du2*dv1 - du1*dv2;

	bump_dv.x = du2*vw1.x - du1*vw2.x;
	bump_dv.y = du2*vw1.y - du1*vw2.y;
	bump_dv.z = du2*vw1.z - du1*vw2.z;

	bump_du.x = dv2*vw1.x - dv1*vw2.x;
	bump_du.y = dv2*vw1.y - dv1*vw2.y;
	bump_du.z = dv2*vw1.z - dv1*vw2.z;
	
	if (d > EPS)
	{
		bump_dv.Normalize();
		bump_du.Normalize();
		bump_dv.Invert();
	}
	else if (d < - EPS)
	{
		bump_dv.Normalize();
		bump_du.Normalize();
		bump_du.Invert();
	}
	else
	{
		bump_dv = Vector(0.0f, 0.0f, 0.0f);
		bump_du = Vector(0.0f, 0.0f, 0.0f);
	}
}

bool Triangle::Intersect(const Vector &o, const Vector &dir, REAL *t)
{
	REAL ndir = ScalarProduct(normalw, dir);
	if (ndir == 0)
		return false;
	REAL no = ScalarProduct(normalw, o);
	REAL tintersect = (planeD - no)/ndir;
	if (tintersect < EPS)
		return false;

	intersectPt.x = o.x + dir.x*tintersect;
	intersectPt.y = o.y + dir.y*tintersect;
	intersectPt.z = o.z + dir.z*tintersect;

	switch (validPlane)
	{
	case axpXY:
		intersectU = ukx*(intersectPt.x - ptw0.x)+ uky*(intersectPt.y - ptw0.y);
		intersectV = vkx*(intersectPt.x - ptw0.x)+ vky*(intersectPt.y - ptw0.y);
		break;
	case axpYZ:
		intersectU = ukx*(intersectPt.y - ptw0.y)+ uky*(intersectPt.z - ptw0.z);
		intersectV = vkx*(intersectPt.y - ptw0.y)+ vky*(intersectPt.z - ptw0.z);
		break;
	case axpZX:
		intersectU = ukx*(intersectPt.z - ptw0.z)+ uky*(intersectPt.x - ptw0.x);
		intersectV = vkx*(intersectPt.z - ptw0.z)+ vky*(intersectPt.x - ptw0.x);
		break;
	default:
		intersectU = - 1;
		intersectV = - 1;
		return false;
	}

	if ((intersectU >= 0) && (intersectV >= 0) && (intersectV + intersectU <= 1))
	{
		*t = tintersect;
		return true;
	}
	return false;
}

void Triangle::PrepareWorldCoord(const Matrix &m)
{
	Transform(&ptw0, pt0, m);
	Transform(&ptw1, pt1, m);
	Transform(&ptw2, pt2, m);

	TransformVector(&normalw0, normal0, m);
	TransformVector(&normalw1, normal1, m);
	TransformVector(&normalw2, normal2, m);

	normalw0.Normalize();
	normalw1.Normalize();
	normalw2.Normalize();
}

void Triangle::GetIntersectNormalw(Vector *normal)
{
	normal->x = normalw0.x + (normalw1.x - normalw0.x)*intersectU + (normalw2.x - normalw0.x)*intersectV;
	normal->y = normalw0.y + (normalw1.y - normalw0.y)*intersectU + (normalw2.y - normalw0.y)*intersectV;
	normal->z = normalw0.z + (normalw1.z - normalw0.z)*intersectU + (normalw2.z - normalw0.z)*intersectV;
	normal->Normalize();
}

void Triangle::GetIntersectTexture(Vector *texture)
{
	texture->x = texture0.x + (texture1.x - texture0.x)*intersectU + (texture2.x - texture0.x)*intersectV;
	texture->y = texture0.y + (texture1.y - texture0.y)*intersectU + (texture2.y - texture0.y)*intersectV;
	texture->z = texture0.z + (texture1.z - texture0.z)*intersectU + (texture2.z - texture0.z)*intersectV;
}
