#include "types.h"


int Min(int a, int b, int c)
{
	if (a < b && a < c)
		return a;
	else if (b < c)
		return b;
	else
		return c;
}
int Max(int a, int b, int c)
{
	if (a > b && a > c)
		return a;
	else if (b > c)
		return b;
	else
		return c;
}


REAL RadianToDegree(REAL value)
{
	return value*(180/PI);
}

REAL DegreeToRadian(REAL value)
{
	return value*(PI/180);
}


void Vector::Normalize()
{
	REAL l = Length();
	if (l > 0.0f)
	{
		x /= l;
		y /= l;
		z /= l;
	}
}


COLORREF Color::GetRGB()
{
	COLORREF color;

	if (b >= 1.0f)
		color = 255;
	else if (b >= 0.0f)
		color = (BYTE)(b*255.0f);
	else
		color = 0;
	color <<= 8;

	if (g >= 1.0f)
		color |= 255;
	else if (g >= 0.0f)
		color |= (BYTE)(g*255.0f);
	color <<= 8;

	if (r >= 1.0f)
		color |= 255;
	else if (r >= 0.0f)
		color |= (BYTE)(r*255.0f);

	return color;
}


REAL Matrix::Minor(int i, int j) const
{
	REAL a[3][3];
	int i0, j0, i1, j1;

	for (i1 = 0, i0 = 0; i1 < 3; i1++, i0++)
	{
		if (i0 == i)
			i0++;
		for (j1 = 0, j0 = 0; j1 < 3; j1++, j0++)
		{
			if (j0 == j)
				j0++;
			a[i1][j1] = m[i0][j0];
		}
	}

	return 
		+ a[0][0]*a[1][1]*a[2][2] + a[0][1]*a[1][2]*a[2][0] + a[0][2]*a[1][0]*a[2][1] 
		- a[0][2]*a[1][1]*a[2][0] - a[0][0]*a[1][2]*a[2][1] - a[0][1]*a[1][0]*a[2][2];
}

REAL Matrix::Determinant() const
{
	return 
		+ _00*_11*_22*_33 - _00*_11*_23*_32 - _00*_12*_21*_33 + _00*_12*_23*_31
		+ _00*_13*_21*_32 - _00*_13*_22*_31 - _01*_10*_22*_33 + _01*_10*_23*_32
		+ _01*_12*_20*_33 - _01*_12*_23*_30 - _01*_13*_20*_32 + _01*_13*_22*_30
		+ _02*_10*_21*_33 - _02*_10*_23*_31 - _02*_11*_20*_33 + _02*_11*_23*_30
		+ _02*_13*_20*_31 - _02*_13*_21*_30 - _03*_10*_21*_32 + _03*_10*_22*_31
		+ _03*_11*_20*_32 - _03*_11*_22*_30 - _03*_12*_20*_31 + _03*_12*_21*_30;
}


void MakeIdentity(Matrix *m)
{
	m->_00 =  1; m->_01 =  0; m->_02 =  0; m->_03 =  0; 
	m->_10 =  0; m->_11 =  1; m->_12 =  0; m->_13 =  0; 
	m->_20 =  0; m->_21 =  0; m->_22 =  1; m->_23 =  0; 
	m->_30 =  0; m->_31 =  0; m->_32 =  0; m->_33 =  1; 
}

void MakeTranslate(Matrix *m, REAL x, REAL y, REAL z)
{
	m->_00 =  1; m->_01 =  0; m->_02 =  0; m->_03 =  0; 
	m->_10 =  0; m->_11 =  1; m->_12 =  0; m->_13 =  0; 
	m->_20 =  0; m->_21 =  0; m->_22 =  1; m->_23 =  0; 
	m->_30 =  x; m->_31 =  y; m->_32 =  z; m->_33 =  1; 
}

void MakeRotateX(Matrix *m, REAL angle)
{
	REAL c = COS(angle);
	REAL s = SIN(angle);

	m->_00 =  1; m->_01 =  0; m->_02 =  0; m->_03 =  0; 
	m->_10 =  0; m->_11 =  c; m->_12 =  s; m->_13 =  0; 
	m->_20 =  0; m->_21 = -s; m->_22 =  c; m->_23 =  0; 
	m->_30 =  0; m->_31 =  0; m->_32 =  0; m->_33 =  1; 
}

void MakeRotateY(Matrix *m, REAL angle)
{
	REAL c = COS(angle);
	REAL s = SIN(angle);

	m->_00 =  c; m->_01 =  0; m->_02 = -s; m->_03 =  0; 
	m->_10 =  0; m->_11 =  1; m->_12 =  0; m->_13 =  0; 
	m->_20 =  s; m->_21 =  0; m->_22 =  c; m->_23 =  0; 
	m->_30 =  0; m->_31 =  0; m->_32 =  0; m->_33 =  1; 
}

void MakeRotateZ(Matrix *m, REAL angle)
{
	REAL c = COS(angle);
	REAL s = SIN(angle);

	m->_00 =  c; m->_01 =  s; m->_02 =  0; m->_03 =  0; 
	m->_10 = -s; m->_11 =  c; m->_12 =  0; m->_13 =  0; 
	m->_20 =  0; m->_21 =  0; m->_22 =  1; m->_23 =  0; 
	m->_30 =  0; m->_31 =  0; m->_32 =  0; m->_33 =  1; 
}

void MakeRotate(Matrix *m, REAL x, REAL y, REAL z, REAL angle)
{
	Quat q;
	MakeQuat(&q, x, y, z, angle);
	MakeMatrix(m, q);
}

void MakeScale(Matrix *m, REAL sx, REAL sy, REAL sz)
{
	m->_00 = sx; m->_01 =  0; m->_02 =  0; m->_03 =  0; 
	m->_10 =  0; m->_11 = sy; m->_12 =  0; m->_13 =  0; 
	m->_20 =  0; m->_21 =  0; m->_22 = sz; m->_23 =  0; 
	m->_30 =  0; m->_31 =  0; m->_32 =  0; m->_33 =  1; 
}

void MakeMatrix(Matrix *res, const Quat &quat)
{
	REAL u, xs, ys, zs, wx, wy, wz, xx, xy, xz, yy, yz, zz;

	// For unit srcQuat, just set s = 2.0; or set xs = i + i, etc.
	REAL lensq = quat.x*quat.x + quat.y*quat.y + quat.z*quat.z + quat.w*quat.w;
	if (lensq > 0)
		u = 2.0f/lensq;
	else
		u = 0;

	xs = quat.x*u;
	ys = quat.y*u;
	zs = quat.z*u;
	wx = quat.w*xs;
	wy = quat.w*ys;
	wz = quat.w*zs;
	xx = quat.x*xs;
	xy = quat.x*ys;
	xz = quat.x*zs;
	yy = quat.y*ys;
	yz = quat.y*zs;
	zz = quat.z*zs;

	res->_00 = 1.0f - (yy + zz);
	res->_01 = xy - wz;
	res->_02 = xz + wy;
	res->_03 = 0.0f;

	res->_10 = xy + wz;
	res->_11 = 1.0f - (xx + zz);
	res->_12 = yz - wx;
	res->_13 = 0.0f;

	res->_20 = xz - wy;
	res->_21 = yz + wx;
	res->_22 = 1.0f - (xx + yy);
	res->_23 = 0.0f;

	res->_30 = 0.0f;
	res->_31 = 0.0f;
	res->_32 = 0.0f;
	res->_33 = 1.0f;
}

void MakeQuat(Quat *res, REAL x, REAL y, REAL z, REAL angle)
{
	Vector axis0 = Vector(x, y, z);
	axis0.Normalize();

	REAL theta, sine;
	theta = 0.5f*angle;
	sine = SIN(theta);

	res->x = axis0.x*sine;
	res->y = axis0.y*sine;
	res->z = axis0.z*sine;
	res->w = COS(theta);
}

void MakeQuat(Quat *res, const Matrix &m);



//    
void MultMatrices(Matrix *res, const Matrix &a, const Matrix &b)
{
	res->_00 = a._00*b._00 + a._01*b._10 + a._02*b._20 + a._03*b._30;
	res->_01 = a._00*b._01 + a._01*b._11 + a._02*b._21 + a._03*b._31;
	res->_02 = a._00*b._02 + a._01*b._12 + a._02*b._22 + a._03*b._32;
	res->_03 = a._00*b._03 + a._01*b._13 + a._02*b._23 + a._03*b._33;

	res->_10 = a._10*b._00 + a._11*b._10 + a._12*b._20 + a._13*b._30;
	res->_11 = a._10*b._01 + a._11*b._11 + a._12*b._21 + a._13*b._31;
	res->_12 = a._10*b._02 + a._11*b._12 + a._12*b._22 + a._13*b._32;
	res->_13 = a._10*b._03 + a._11*b._13 + a._12*b._23 + a._13*b._33;

	res->_20 = a._20*b._00 + a._21*b._10 + a._22*b._20 + a._23*b._30;
	res->_21 = a._20*b._01 + a._21*b._11 + a._22*b._21 + a._23*b._31;
	res->_22 = a._20*b._02 + a._21*b._12 + a._22*b._22 + a._23*b._32;
	res->_23 = a._20*b._03 + a._21*b._13 + a._22*b._23 + a._23*b._33;

	res->_30 = a._30*b._00 + a._31*b._10 + a._32*b._20 + a._33*b._30;
	res->_31 = a._30*b._01 + a._31*b._11 + a._32*b._21 + a._33*b._31;
	res->_32 = a._30*b._02 + a._31*b._12 + a._32*b._22 + a._33*b._32;
	res->_33 = a._30*b._03 + a._31*b._13 + a._32*b._23 + a._33*b._33;
}

void MultMatrixLeft(Matrix *res, const Matrix &m)
{
	Matrix orig = *res;
	MultMatrices(res, m, orig);
}

void MultMatrixRight(Matrix *res, const Matrix &m)
{
	Matrix orig = *res;
	MultMatrices(res, orig, m);
}

void Translate(Matrix *m, REAL x, REAL y, REAL z)
{
	Matrix multM;
	MakeTranslate(&multM, x, y, z);
	MultMatrixRight(m, multM);
}

void RotateX(Matrix *m, REAL angle)
{
	Matrix multM;
	MakeRotateX(&multM, angle);
	MultMatrixRight(m, multM);
}

void RotateY(Matrix *m, REAL angle)
{
	Matrix multM;
	MakeRotateY(&multM, angle);
	MultMatrixRight(m, multM);
}

void RotateZ(Matrix *m, REAL angle)
{
	Matrix multM;
	MakeRotateZ(&multM, angle);
	MultMatrixRight(m, multM);
}

void Rotate(Matrix *m, REAL x, REAL y, REAL z, REAL angle)
{
	Matrix multM;
	MakeRotate(&multM, x, y, z, angle);
	MultMatrixRight(m, multM);
}

void Scale(Matrix *m, REAL sx, REAL sy, REAL sz)
{
	Matrix multM;
	MakeScale(&multM, sx, sy, sz);
	MultMatrixRight(m, multM);
}


/*void ReverseMatrix(Matrix *res, const Matrix &m)
{
	REAL d = m.Determinant(), k;
	if (d == 0)
		throw "Matrix invalid";
	k = 1/d;
	for (int i = 0; i < 4; i++)
		for (int j = 0; j < 4; j++)
			if ((i + j)%2 == 0)
				res->m[j][i] = m.Minor(i, j)*k;
			else
				res->m[j][i] = - m.Minor(i, j)*k;
}*/

void ReverseMatrix(Matrix *res, const Matrix &m)
{
	REAL d = m.Determinant(), k;
	if (d == 0)
		throw L"Matrix invalid";
	k = 1/d;

	res->_00 = (m._11*m._22*m._33 + m._12*m._23*m._31 + m._13*m._21*m._32 - m._13*m._22*m._31 - m._11*m._23*m._32 - m._12*m._21*m._33)*k;
	res->_10 = (m._13*m._22*m._30 + m._10*m._23*m._32 + m._12*m._20*m._33 - m._10*m._22*m._33 - m._12*m._23*m._30 - m._13*m._20*m._32)*k;
	res->_20 = (m._10*m._21*m._33 + m._11*m._23*m._30 + m._13*m._20*m._31 - m._13*m._21*m._30 - m._10*m._23*m._31 - m._11*m._20*m._33)*k;
	res->_30 = (m._12*m._21*m._30 + m._10*m._22*m._31 + m._11*m._20*m._32 - m._10*m._21*m._32 - m._11*m._22*m._30 - m._12*m._20*m._31)*k;
	res->_01 = (m._03*m._22*m._31 + m._01*m._23*m._32 + m._02*m._21*m._33 - m._01*m._22*m._33 - m._02*m._23*m._31 - m._03*m._21*m._32)*k;
	res->_11 = (m._00*m._22*m._33 + m._02*m._23*m._30 + m._03*m._20*m._32 - m._03*m._22*m._30 - m._00*m._23*m._32 - m._02*m._20*m._33)*k;
	res->_21 = (m._03*m._21*m._30 + m._00*m._23*m._31 + m._01*m._20*m._33 - m._00*m._21*m._33 - m._01*m._23*m._30 - m._03*m._20*m._31)*k;
	res->_31 = (m._00*m._21*m._32 + m._01*m._22*m._30 + m._02*m._20*m._31 - m._02*m._21*m._30 - m._00*m._22*m._31 - m._01*m._20*m._32)*k;
	res->_02 = (m._01*m._12*m._33 + m._02*m._13*m._31 + m._03*m._11*m._32 - m._03*m._12*m._31 - m._01*m._13*m._32 - m._02*m._11*m._33)*k;
	res->_12 = (m._03*m._12*m._30 + m._00*m._13*m._32 + m._02*m._10*m._33 - m._00*m._12*m._33 - m._02*m._13*m._30 - m._03*m._10*m._32)*k;
	res->_22 = (m._00*m._11*m._33 + m._01*m._13*m._30 + m._03*m._10*m._31 - m._03*m._11*m._30 - m._00*m._13*m._31 - m._01*m._10*m._33)*k;
	res->_32 = (m._02*m._11*m._30 + m._00*m._12*m._31 + m._01*m._10*m._32 - m._00*m._11*m._32 - m._01*m._12*m._30 - m._02*m._10*m._31)*k;
	res->_03 = (m._03*m._12*m._21 + m._01*m._13*m._22 + m._02*m._11*m._23 - m._01*m._12*m._23 - m._02*m._13*m._21 - m._03*m._11*m._22)*k;
	res->_13 = (m._00*m._12*m._23 + m._02*m._13*m._20 + m._03*m._10*m._22 - m._03*m._12*m._20 - m._00*m._13*m._22 - m._02*m._10*m._23)*k;
	res->_23 = (m._03*m._11*m._20 + m._00*m._13*m._21 + m._01*m._10*m._23 - m._00*m._11*m._23 - m._01*m._13*m._20 - m._03*m._10*m._21)*k;
	res->_33 = (m._00*m._11*m._22 + m._01*m._12*m._20 + m._02*m._10*m._21 - m._02*m._11*m._20 - m._00*m._12*m._21 - m._01*m._10*m._22)*k;
}

void VectorAddition(Vector *res, const Vector &a, const Vector &b)
{
	res->x = a.x + b.x;
	res->y = a.y + b.y;
	res->z = a.z + b.z;
}

void VectorSubtraction(Vector *res, const Vector &a, const Vector &b)
{
	res->x = a.x - b.x;
	res->y = a.y - b.y;
	res->z = a.z - b.z;
}

void VectorScale(Vector *res, const Vector &v, REAL k)
{
	res->x = v.x*k;
	res->y = v.y*k;
	res->z = v.z*k;
}

void VectorProduct(Vector *res, const Vector &a, const Vector &b)
{
	res->x = a.y*b.z - a.z*b.y;
	res->y = a.z*b.x - a.x*b.z;
	res->z = a.x*b.y - a.y*b.x;
}

REAL ScalarProduct(const Vector &a, const Vector &b)
{
	return a.x*b.x + a.y*b.y + a.z*b.z;
}

REAL ScalarProductNorm(const Vector &a, const Vector &b)
{
	REAL lenABSq = a.LengthSq()*b.LengthSq();
	if (lenABSq > 0)
		return (a.x*b.x + a.y*b.y + a.z*b.z)/SQRT(lenABSq);
	else
		return 0;
}


//     (v = v*m)
void Transform(Vector *res, const Vector &v, const Matrix &m)
{
	res->x = m._00*v.x + m._10*v.y + m._20*v.z + m._30;
	res->y = m._01*v.x + m._11*v.y + m._21*v.z + m._31;
	res->z = m._02*v.x + m._12*v.y + m._22*v.z + m._32;
}

//     (v = v*m)
void TransformVector(Vector *res, const Vector &v, const Matrix &m)
{
	res->x = m._00*v.x + m._10*v.y + m._20*v.z;
	res->y = m._01*v.x + m._11*v.y + m._21*v.z;
	res->z = m._02*v.x + m._12*v.y + m._22*v.z;
}

REAL Randomf(REAL min, REAL max)
{
	return (min + (rand()%1000)*(max - min)/1000);
}
