/******************************************************************************
*                                                                             *
*                                                    *
*                                                     *
*                                                                             *
*   15  2004 .                                                        *
*   (C)                                                         *
*   e-mail       pavelkok@mail.ru                                             *
*   home page    http://kps.nm.ru                                             *
*                                                                             *
*                                                                             *
*   types.h                                                                   *
*    ,                                                   *
*    Color, Vector, Quat, Matrix                                   *
*                                                             *
*                                                                             *
*******************************************************************************/


#ifndef TYPES_H
#define TYPES_H


#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#define UNICODE
#define _WIN32_WINNT					0x0500
#define WINVER							0x0500
#include <windows.h>
#include <commctrl.h>

#include "array.h"
#include "string.h"


// types
typedef unsigned char byte;
typedef unsigned short word;
typedef unsigned int uint;
typedef unsigned long ulong;
typedef unsigned char BYTE;
typedef unsigned long COLORREF;
typedef COLORREF *PCOLORREF;

typedef wchar_t wchar;


#ifdef REAL64

	#define REAL							double
	#define SQRT							sqrt
	#define SIN								sin
	#define COS								cos
	#define FABS							fabs
	#define FLOOR							floor
	#define POW								pow

	#define INFINITY						1e10
	#define EPS								1e-8
	#define EPS_BOUNDING					0.1
	#define PI								3.14159265359

#endif
#ifdef REAL32

	#define REAL							float
	#define SQRT							sqrtf
	#define SIN								sinf
	#define COS								cosf
	#define FABS							fabsf
	#define FLOOR							floorf
	#define POW								powf

	#define INFINITY						1e10f
	#define EPS								1e-3f
	#define EPS_BOUNDING					0.1f
	#define PI								3.14159265359f

#endif


enum AxisPlane
{
	axpXY = 1,
	axpYZ = 2,
	axpZX = 3
};


typedef class Bitmap *PBitmap;			//   + HBITMAP
typedef class Triangle *PTriangle;
typedef class Node *PNode;				//  
typedef class Group *PGroup;			//  PNode
typedef class Mesh *PMesh;				//  
typedef class Light *PLight;			// . 
typedef class Camera *PCamera;			// 
typedef class Texture *PTexture;
typedef class Material *PMaterial;
typedef class LoaderASE *PLoaderASE;	//   3dsmax  ASCII Scene
typedef class Scene *PScene;			//    

typedef SimpleArray<PTriangle> SimpleTriangleArray, *PSimpleTriangleArray;
typedef Array<PTriangle> TriangleArray, *PTriangleArray;
typedef Array<PNode> NodeArray, *PNodeArray;
typedef Array<PLight> LightArray, *PLightArray;
typedef Array<PCamera> CameraArray, *PCameraArray;
typedef Array<PTexture> TextureArray, *PTextureArray;
typedef Array<PMaterial> MaterialArray, *PMaterialArray;


struct Color
{
	REAL r;
	REAL g;
	REAL b;

	Color() { r = g = b = 0; }
	Color(REAL r, REAL g, REAL b) { this->r = r; this->g = g; this->b = b; }

	void Add(const Color &c) { r += c.r; g += c.g; b += c.b; }
	void Invert() { r = 1.0f - r; g = 1.0f - g; b = 1.0f - b; }

	bool IsBlack() { return (r == 0.0f) && (g == 0.0f) && (b == 0.0f); }
	bool IsWhite() { return (r == 1.0f) && (g == 1.0f) && (b == 1.0f); }

	COLORREF GetRGB();
	REAL GetIntens() { return (r + g + b)/3.0f; }
	REAL GetAmountDiff(const Color &c) { return FABS(r - c.r) + FABS(g - c.g) + FABS(b - c.b); }
};

struct Vector
{
	REAL x;
	REAL y;
	REAL z;

	Vector() { x = y = z = 0; }
	Vector(REAL x, REAL y, REAL z) { this->x = x; this->y = y; this->z = z; }

	REAL Length() const { return SQRT(x*x + y*y + z*z); }
	REAL LengthSq() const { return (x*x + y*y + z*z); }
	void Normalize();
	void Invert() { x = - x; y = - y; z = - z; }
};

struct Quat
{
	REAL x;
	REAL y;
	REAL z;
	REAL w;

	Quat() { x = y = z = 0; w = 1; }
	Quat(REAL x, REAL y, REAL z, REAL w) { this->x = x; this->y = y; this->z = z; this->w = w; }
};

struct Matrix
{
	union
	{
		struct
		{
			REAL _00, _01, _02, _03;
			REAL _10, _11, _12, _13;
			REAL _20, _21, _22, _23;
			REAL _30, _31, _32, _33;
		};
		REAL m[4][4];
	};

	Matrix()
	{
		_00 = _01 = _02 = _03 = 0;
		_10 = _11 = _12 = _13 = 0;
		_20 = _21 = _22 = _23 = 0;
		_30 = _31 = _32 = _33 = 0;
	}

	Matrix(
		REAL _00, REAL _01, REAL _02, REAL _03, 
		REAL _10, REAL _11, REAL _12, REAL _13, 
		REAL _20, REAL _21, REAL _22, REAL _23, 
		REAL _30, REAL _31, REAL _32, REAL _33)
	{
		this->_00 = _00; this->_01 = _01; this->_02 = _02; this->_03 = _03; 
		this->_10 = _10; this->_11 = _11; this->_12 = _12; this->_13 = _13; 
		this->_20 = _20; this->_21 = _21; this->_22 = _22; this->_23 = _23; 
		this->_30 = _30; this->_31 = _31; this->_32 = _32; this->_33 = _33; 
	}

	REAL Minor(int i, int j) const;
	REAL Determinant() const;
};


int Min(int a, int b, int c);
int Max(int a, int b, int c);
int Min(int a, int b);
int Max(int a, int b);

REAL Randomf(REAL min, REAL max);

REAL RadianToDegree(REAL value);
REAL DegreeToRadian(REAL value);

void MakeIdentity(Matrix *m);
void MakeTranslate(Matrix *m, REAL x, REAL y, REAL z);
void MakeRotateX(Matrix *m, REAL angle);
void MakeRotateY(Matrix *m, REAL angle);
void MakeRotateZ(Matrix *m, REAL angle);
void MakeRotate(Matrix *m, REAL x, REAL y, REAL z, REAL angle);
void MakeScale(Matrix *m, REAL sx, REAL sy, REAL sz);

void MakeMatrix(Matrix *res, const Quat &quat);
void MakeQuat(Quat *res, REAL x, REAL y, REAL z, REAL angle);
void MakeQuat(Quat *res, const Matrix &m);

void Translate(Matrix *m, REAL x, REAL y, REAL z);
void RotateX(Matrix *m, REAL angle);
void RotateY(Matrix *m, REAL angle);
void RotateZ(Matrix *m, REAL angle);
void Rotate(Matrix *m, REAL x, REAL y, REAL z, REAL angle);
void Scale(Matrix *m, REAL sx, REAL sy, REAL sz);

void MultMatrices(Matrix *res, const Matrix &a, const Matrix &b);
void MultMatrixLeft(Matrix *res, const Matrix &m);
void MultMatrixRight(Matrix *res, const Matrix &m);

void ReverseMatrix(Matrix *res, const Matrix &m);

void VectorAddition(Vector *res, const Vector &a, const Vector &b);
void VectorSubtraction(Vector *res, const Vector &a, const Vector &b);
void VectorScale(Vector *res, const Vector &v, REAL k);
void VectorProduct(Vector *res, const Vector &a, const Vector &b);
REAL ScalarProduct(const Vector &a, const Vector &b);
REAL ScalarProductNorm(const Vector &a, const Vector &b);

void Transform(Vector *res, const Vector &v, const Matrix &m);
void TransformVector(Vector *res, const Vector &v, const Matrix &m);

#endif
